#!/usr/bin/python3.6
# -*- coding:UTF-8 -*-
import numpy as np
import os
import pandas as pd
import meteva
import traceback
import re
import copy
from . import DataBlock_pb2
from .GDS_data_service import GDSDataService
import struct
from collections import OrderedDict
import datetime
import math
from io import StringIO
from .CMADaasAccess import CMADaasAccess
from .httpclient import get_http_result_cimiss
import json


def read_station(filename,keep_alt = False,show = False):
    '''
    :param filename: 站点文件路径，它可以是micaps第1、2、3、8类文件
    :return: 站点数据，其中time,dtime,level属性为设置的缺省值，数据内容都设置为0
    '''
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding,_ = meteva.base.io.get_encoding_of_file(filename,read_rows=1)
        if encoding is None:
            print("文件编码格式不识别")
            return None
        try:
            file = open(filename, encoding=encoding)
            sta = None
            head = file.readline()
            strs = head.split()
            if strs[1] == "3":
                if keep_alt:
                    sta = read_sta_alt_from_micaps3(filename)
                else:
                    sta = read_stadata_from_micaps3(filename)
            elif strs[1] == "16":
                sta = read_stadata_from_micaps16(filename)
            elif strs[1] == "1" or str[1] == "2" or str[1] == "8":
                sta = read_stadata_from_micaps1_2_8(filename,column=3)
            else:
                print(filename + "is not micaps第1、2、3、8类文件")

            if sta is not None:
                data_name = sta.columns[-1]
                if not keep_alt:
                    sta[data_name] = 0
                else:
                    meteva.base.set_stadata_names(sta,["alt"])
                meteva.base.set_stadata_coords(sta,time = datetime.datetime(2099,1,1,8,0),level = 0,dtime= 0)
                if show:
                    print("success read from "+filename)
                return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)

            print(filename + "文件格式不能识别。可能原因：文件未按micaps格式存储")
            return None

def read_sta_alt_from_micaps3(filename, station=None, drop_same_id=True,show = False):
    '''
    读取micaps3格式文件转换为pandas中dataframe结构的数据

    :param reserve_time_dtime_level:保留时间，时效和层次，默认为rue
    :param data_name:dataframe中数值的values列的名称
    :return:返回一个dataframe结构的多列站点数据。
    :param filename: 文件路径
    :param station: 站号，默认：None
    :param drop_same_id: 是否要删除相同id的行  默认为True
    :return:
    '''

    #print(os.path.exists(filename))
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding, _ = meteva.base.io.get_encoding_of_file(filename,read_rows=1)
        if encoding is None:
            print("文件编码格式不识别")
            return None

        try:
            file = open(filename, 'r',encoding= encoding)
            skip_num = 0
            strs = []
            nline = 0
            nregion = 0
            nstart = 0
            while 1 > 0:
                skip_num += 1
                str1 = file.readline()
                strs.extend(str1.split())

                if (len(strs) > 8):
                    nline = int(strs[8])
                if (len(strs) > 11 + nline):
                    nregion = int(strs[11 + nline])
                    nstart = nline + 2 * nregion + 14
                    if (len(strs) == nstart):
                        break
            file.close()

            file_sta = open(filename,'r',encoding= encoding)

            sta1 = pd.read_csv(file_sta, skiprows=skip_num, sep="\s+", header=None, usecols=[0, 1, 2, 3])
            sta1.columns = ['id', 'lon', 'lat', 'alt']
            sta1.drop_duplicates(keep='first', inplace=True)
            if drop_same_id:
                sta1 = sta1.drop_duplicates(['id'])
            # sta = bd.sta_data(sta1)
            sta = meteva.base.basicdata.sta_data(sta1)
            # print(sta)

            y2 = ""
            if len(strs[3]) == 2:
                year = int(strs[3])
                if year >= 50:
                    y2 = '19'
                else:
                    y2 = '20'
            if len(strs[3]) == 1: strs[3] = "0" + strs[3]
            if len(strs[4]) == 1: strs[4] = "0" + strs[4]
            if len(strs[5]) == 1: strs[5] = "0" + strs[5]
            if len(strs[6]) == 1: strs[6] = "0" + strs[6]

            time_str = y2 + strs[3] + strs[4] + strs[5] + strs[6]
            time_file = meteva.base.tool.time_tools.str_to_time(time_str)
            sta.loc[:,"time"] = time_file
            sta.loc[:,"dtime"] = 0
            sta.loc[:,"level"] = 0 #int(strs[7])

            if (station is not None):
                sta = meteva.base.put_stadata_on_station(sta, station)
            if show:
                print("success read from " + filename)
            return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)

            print(filename+"文件格式不能识别。可能原因：文件未按micaps3格式存储")
            return None

def read_stadata_from_micaps3(filename, station=None,  level=None,time=None, dtime=None, data_name='data0', drop_same_id=True,show = False):
    '''
    读取micaps3格式文件转换为pandas中dataframe结构的数据

    :param reserve_time_dtime_level:保留时间，时效和层次，默认为rue
    :param data_name:dataframe中数值的values列的名称
    :return:返回一个dataframe结构的多列站点数据。
    :param filename: 文件路径
    :param station: 站号，默认：None
    :param time: 起报时  默认：NOne
    :param dtime: 时效 默认：None
    :param level:  层次  默认：None
    :param data_name: 要素名  默认：'data0'
    :param drop_same_id: 是否要删除相同id的行  默认为True
    :return:
    '''
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding, _ = meteva.base.io.get_encoding_of_file(filename,read_rows=1)
        if encoding is None:
            print("文件编码格式不识别")
            return None

        try:
            file = open(filename, 'r',encoding=encoding)
            skip_num = 0
            strs = []
            nline = 0
            nregion = 0
            nstart = 0
            while 1 > 0:
                skip_num += 1
                str1 = file.readline()
                #print(str1)
                strs.extend(str1.split())

                if (len(strs) > 8):
                    nline = int(strs[8])
                if (len(strs) > 11 + nline):
                    nregion = int(strs[11 + nline])
                    nstart = nline + 2 * nregion + 14
                    if (len(strs) == nstart):
                        break
            file.close()
            #print(skip_num)
            if int(strs[-1]) == 0:return None

            file_sta = open(filename, 'r',encoding=encoding)
            sta1 = pd.read_csv(file_sta, skiprows=skip_num, sep="\s+", header=None, usecols=[0, 1, 2, 4])
            sta1.columns = ['id', 'lon', 'lat', data_name]
            sta1.drop_duplicates(keep='first', inplace=True)
            if drop_same_id:
                sta1 = sta1.drop_duplicates(['id'])
            # sta = bd.sta_data(sta1)
            sta = meteva.base.basicdata.sta_data(sta1)
            #print(sta)

            y2 = ""
            if len(strs[3]) == 2:
                year = int(strs[3])
                if year >= 50:
                    y2 = '19'
                else:
                    y2 = '20'
            if len(strs[3]) == 1: strs[3] = "0" + strs[3]
            if len(strs[4]) == 1: strs[4] = "0" + strs[4]
            if len(strs[5]) == 1: strs[5] = "0" + strs[5]
            if len(strs[6]) == 1: strs[6] = "0" + strs[6]

            if time is None:
                time_str = y2 + strs[3] + strs[4] + strs[5] + strs[6]
                time_file = meteva.base.tool.time_tools.str_to_time(time_str)
                sta.loc[:,"time"] = time_file
            else:
                time_file = meteva.base.tool.time_tools.all_type_time_to_datetime(time)
                sta.loc[:,"time"] = time_file
            sta.loc[:,"dtime"] = 0
            sta.loc[:,"level"] = 0 #int(strs[7])
            #print(time_str)
            meteva.base.set_stadata_coords(sta, level=level, dtime=dtime)

            if (station is not None):
                sta = meteva.base.put_stadata_on_station(sta, station)
            if show:
                print("success read from " + filename)
            return sta

        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)

            print(filename+"文件格式不能识别。可能原因：文件未按micaps3格式存储")
            return None

def read_stadata_from_csv(filename, columns, member_list,skiprows=0,level = None,time = None,dtime = None, drop_same_id=False,show = False,
                          sep = "\s+"):

    """
    读取站点数据
    :param filename:带有站点信息的路径已经文件名
    :param columns 列名
    :param skiprows:读取时跳过的行数，默认为：0
    :param drop_same_id: 是否要删除相同id的行  默认为True
    :return:返回带有'level','time','dtime','id','lon','lat','alt','data0'列的dataframe站点信息。
    """
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding,_ = meteva.base.io.get_encoding_of_file(filename,read_rows=skiprows)
        if encoding is None:
            print("文件编码格式不识别")
            return None
        try:
            file_sta = open(filename, 'r',encoding = encoding)
            if "time" in columns:
                index = columns.index("time")
                sta0 = pd.read_csv(file_sta, skiprows=skiprows,  header=None,sep=sep,parse_dates=[index])
                sta0.columns = columns
            else:
                sta0 = pd.read_csv(file_sta, skiprows=skiprows, header=None, sep=sep)
                sta0.columns = columns

            station_column = []
            for column in columns:
                if column in meteva.base.get_coord_names():
                    station_column.append(column)
            if not isinstance(member_list,list):
                member_list = [member_list]

            sta0.drop_duplicates(keep='first', inplace=True)
            station_column.extend(member_list)
            sta1 = sta0[station_column]
            nsta = len(sta1.index)
            if "lon" in station_column:
                if sta1.loc[0,"lon"] >1000:
                    a = sta1.loc[:, 'lon'] // 100 + (sta1.loc[:, 'lon'] %100)/60
                    sta1.loc[:, "lon"] = a
            if "lat" in station_column:
                if sta1.loc[0, "lat"] > 1000:
                    a = sta1.loc[:, 'lat'] // 100 + (sta1.loc[:, 'lat'] % 100) / 60
                    sta1.loc[:, "lat"] = a
                # for i in range(nsta):
                #     if sta1.loc[i, 'lon'] > 1000:
                #         a = sta1.loc[i, 'lon'] // 100 + (a % 100) / 60
                #         sta1.loc[i, 'lon'] = a
            # if "lat" in station_column:
            #     for i in range(nsta):
            #         if sta1.loc[i, 'lat'] > 1000:
            #             a = sta1.loc[i, 'lat'] // 100 + (a % 100) / 60
            #             sta1.loc[i, 'lat'] = a
            # sta = bd.sta_data(sta1)
            sta = meteva.base.basicdata.sta_data(sta1)
            if drop_same_id:
                sta = sta.drop_duplicates(['id'])

            # sta['time'] = method.time_tools.str_to_time64("2099010108")
            if time is not None:
                sta.loc[:,"time"] = time
            elif pd.isnull(sta.iloc[0,1]):
                sta.loc[:,'time'] = meteva.base.tool.time_tools.str_to_time64("2099010108")
            if dtime is not None:
                sta.loc[:,"dtime"] = dtime
            elif pd.isnull(sta.iloc[0,2]):
                sta.loc[:,"dtime"] = 0

            if level is not None:
                sta.loc[:,"level"] = level
            elif pd.isnull(sta.iloc[0,0]):
                sta.loc[:,"level"] = 0


            meteva.base.basicdata.reset_id(sta)
            if show:
                print("success read from "+filename)
            return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename+"文件格式不能识别。可能原因：文件未按pandas能够识别的csv格式存储")
            return None



def read_stadata_from_sevp(filename, element_id,level=None,time=None,data_name = "data0",show = False):
    '''
    兼容多个时次的预报产品文件 txt格式
    :param：filename:文件路径和名称
    :param: element:选取要素
    :param drop_same_id: 是否要删除相同id的行  默认为True
    :return：dataframe格式的站点数据

    '''

    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding,lines = meteva.base.io.get_encoding_of_file(filename,read_rows=6)
        if encoding is None:
            print("文件编码格式不识别")
            return None
        try:
            #lines = heads.split("\n")
            file = open(filename,encoding = encoding)
            str_lines0 = file.readlines()
            str_lines = []
            for str1 in str_lines0:
                if str1.strip() !="":
                    str_lines.append(str1)
            if time is None:
                strs4 = lines[3].split()
                time_file = meteva.base.tool.time_tools.str_to_time(strs4[1])
            else:
                time_file = time
            if level is None:
                level = 0

            num_all = len(str_lines) - 1
            str_lines_list = []
            nline = 5
            while nline< num_all:
                str1 = str_lines[nline][:-1] +" "
                strs = str1.split()
                strs_len = len(strs)
                if(strs_len>5):
                    next_num = int(strs[4])
                    if strs_len > 6:
                        str1 = strs[0]+" " + strs[1] +" "+ strs[2] +" "+ strs[3] +" "+ strs[4] +" "+ strs[5] + " "
                    for i in range(1,next_num+1):
                        e_str = str1 + str_lines[nline+i]
                        str_lines_list.append(e_str)
                    nline += next_num+1
                else:
                    nline +=  1
            strs_all = "".join(str_lines_list)

            f = StringIO(strs_all)
            df = pd.read_csv(f,header = None,sep="\s+")
            df = df.loc[:,[0,1,2,6,6 + element_id]]
            df.columns = ["id", "lon", "lat", "dtime",data_name]
            df["time"] = time_file
            df["level"] = level
            sta = meteva.base.sta_data(df)
            return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename + " 文件格式异常")


def read_stadata_from_micaps1_2_8(filename, column, station=None, level=None,time=None, dtime=None, data_name='data0', drop_same_id=True,show = False):
    '''
    read_from_micaps1_2_8  读取m1、m2、m8格式的文件
    :param filename: 文件路径
    :param column: 选取哪列要素  4-len
    :param station: 站号 默认为None
    :param drop_same_id: 是否要删除相同id的行  默认为True
    :return:
    '''
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding,heads = meteva.base.io.get_encoding_of_file(filename,read_rows=2)
        if encoding is None:
            print("文件编码格式不识别")
            return None

        try:
            file = open(filename,encoding=encoding)
            sta1 = pd.read_csv(file, skiprows=2, sep="\s+", header=None, usecols=[0, 1, 2,  column])
            sta1.columns = ['id', 'lon', 'lat', data_name]
            sta2 = meteva.base.basicdata.sta_data(sta1)
            if drop_same_id:
                sta2 = sta2.drop_duplicates(['id'])
            strs0 = heads[0].split()
            strs = heads[1].split()
            y2 = ""
            if len(strs[0]) == 2:
                year = int(strs[0])
                if year >= 50:
                    y2 = '19'
                else:
                    y2 = '20'
            if len(strs[0]) == 1: strs[0] = "0" + strs[0]
            if len(strs[1]) == 1: strs[1] = "0" + strs[1]
            if len(strs[2]) == 1: strs[2] = "0" + strs[2]
            if len(strs[3]) == 1: strs[3] = "0" + strs[3]

            time_str = y2 + strs[0] + strs[1] + strs[2] + strs[3]
            time_file = meteva.base.tool.time_tools.str_to_time(time_str)
            if time is None:
                sta2.loc[:,'time'] = time_file
            else:
                sta2.loc[:,'time'] = time
            #print(strs0)
            if strs0[1] == "1":
                sta2.loc[:,"level"] = 0
                sta2.loc[:,"dtime"] = 0
            elif strs0[1] == "2":
                sta2.loc[:,"level"] = int(strs[4])
                sta2.loc[:,"dtime"] = 0
            elif strs0[1] == "8":
                sta2.loc[:,"level"] = 0
                sta2.loc[:,"dtime"] = int(strs[4])
            else:
                print(filename + "is not micaps第1、2、3、8类文件")

            meteva.base.set_stadata_coords(sta2,level= level,time = time,dtime= dtime)
            if show:
                print("success read from "+filename)
            if station is None:
                return sta2
            else:
                sta = meteva.base.put_stadata_on_station(sta2, station)
                return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename+"文件格式不能识别。可能原因：文件未按micaps第1、2、8类格式存储")
            return None


def read_stadata_from_micaps41_lightning(filename, column, level=0,data_name='data0', show = False,keep_millisecond = False):
    '''
        read_from_micaps41  读取m41格式的文件
        :param filename: 文件路径
        :param column: 选取哪列要素  4-len
        :param station: 站号 默认为None
        :param drop_same_id: 是否要删除相同id的行  默认为True
        :return:
        '''
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding, heads = meteva.base.io.get_encoding_of_file(filename, read_rows=2)
        if encoding is None:
            print("文件编码格式不识别")
            return None

        try:
            file = open(filename, encoding=encoding)
            sta1 = pd.read_csv(file, skiprows=2, sep="\s+", header=None, usecols=[0, 1, 4,5,column])
            sta1.columns = ['id','time', 'lon', 'lat', data_name]
            sta1['time'] = pd.to_datetime(sta1['time'],format="%Y%m%d%H%M%S%f")
            if not keep_millisecond:
                sta1['time'] = sta1['time'].dt.ceil("S")
            sta2 = meteva.base.basicdata.sta_data(sta1)

            sta2.loc[:, "level"] = level
            sta2.loc[:, "dtime"] = 0

            if show:
                print("success read from " + filename)
            return sta2

        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename + "文件格式不能识别。可能原因：文件未按micaps第41类格式存储")
            return None


def set_io_config(filename):
    if os.path.exists(filename):
        try:
            ip,port = read_gds_ip_port(filename)
            ips = ip.split(".")
            if len(ips) == 4:
                for i in range(4):
                    digital = int(ips[i])
            else:
                print("filename  内容的格式不符合要求")
            meteva.base.gds_ip_port = ip,port
            print("配置文件设置成功")
        except:
            print(filename + "中micaps分布式数据库设置格式不符合要求")

        try:
            meteva.base.cimiss_set = read_cimiss_set(filename)
        except:
            print(filename + "中未包含cimiss数据库设置，或格式不符合要求，不读取cmiss数据库中数据则可忽略该信息")

        try:
            meteva.base.cmadaas_set = read_cmadass_set(filename)
        except:
            print(filename + "中未包含cmadaas数据库设置，或格式不符合要求，不读取大数据云平台则可忽略该信息")
    else:
        print("cimiss和gds配置文件不存在")


def read_cimiss_set(filename,show = False):
    if filename is None:
        print("请使用set_config_ip_port 函数设置存储ip，port的配置文件的路径")
    file = open(filename)
    for i in range(100):
        title = file.readline()
        if title.find("CIMISS") >=0:
            break
    dns = file.readline().split("=")[1]
    dns = dns.strip()
    userId =file.readline().split("=")[1]
    userId = userId.strip()
    pwd = file.readline().split("=")[1]
    pwd = pwd.strip()
    file.close()
    if show:
        print("success read from " + filename)
    return dns,userId,pwd

def read_cmadass_set(filename,show = False):
    if filename is None:
        print("请使用set_config_ip_port 函数设置存储ip，port的配置文件的路径")
    file = open(filename)
    for i in range(100):
        title = file.readline()
        if title.find("CMADaaS") >=0:
            break
    dns = file.readline().split("=")[1]
    dns = dns.strip()
    port = file.readline().split("=")[1]
    port = port.strip()
    userId =file.readline().split("=")[1]
    userId = userId.strip()
    pwd = file.readline().split("=")[1]
    pwd = pwd.strip()
    file.close()
    if show:
        print("success read from " + filename)
    return dns,port,userId,pwd

def read_gds_ip_port(filename,show = False):
    if filename is None:
        print("请使用set_config_ip_port 函数设置存储ip，port的配置文件的路径")
    file = open(filename)
    for i in range(100):
        title = file.readline()
        if title.find("MICAPS") >=0:
            break
    ip = file.readline().split("=")[1]
    ip = ip.strip()
    port = int(file.readline().split("=")[1])
    file.close()
    if show:
        print("success read from " + filename)
    return ip,port

def read_stadata_from_gds(filename,element_id = None,station = None, level=None,time=None, dtime=None, data_name='data0',show = False):
    '''
    :param ip: 为字符串形式，示例 “10.20.30.40”
    :param port: 为整数形式 示例 8080
    :param filename0:
    :param element_id0:
    :param station:
    :param data_name:
    :return:
    '''
    directory, filename = os.path.split(filename)
    # connect to data service
    if meteva.base.gds_ip_port is None:
        print("请先使用set_config 配置gds的ip和port")
        return
    ip,port = meteva.base.gds_ip_port
    service = GDSDataService(ip, port)

    # get data file name
    element_id0 = element_id
    if element_id is not None:
        element_id_str0 = str(element_id)

    try:
        directory = directory.replace("mdfs:///", "")
        directory = directory.replace("\\","/")
        status, response = service.getData(directory, filename)
    except ValueError:
        print('Can not retrieve data' + filename + ' from ' + directory)
        return None
    try:
        ByteArrayResult = DataBlock_pb2.ByteArrayResult()
        if status == 200:
            ByteArrayResult.ParseFromString(response)
            if ByteArrayResult is not None:
                byteArray = ByteArrayResult.byteArray

                # define head structure
                head_dtype = [('discriminator', 'S4'), ('type', 'i2'),
                              ('description', 'S100'),
                              ('level', 'f4'), ('levelDescription', 'S50'),
                              ('year', 'i4'), ('month', 'i4'), ('day', 'i4'),
                              ('hour', 'i4'), ('minute', 'i4'), ('second', 'i4'),
                              ('Timezone', 'i4'),("id_type","i2"), ('extent', 'S98')]

                # read head information
                head_info = np.frombuffer(byteArray[0:288], dtype=head_dtype)
                #print(head_info)
                if time is None:
                    time = datetime.datetime(
                        head_info['year'][0], head_info['month'][0],
                        head_info['day'][0], head_info['hour'][0],
                        head_info['minute'][0], head_info['second'][0])
                else:
                    time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
                if level is None:
                    level = head_info["level"][0]
                if dtime is None:
                    filename1 = os.path.split(filename)[1].split(".")
                    dtime = int(filename1[1])
                id_type = head_info['id_type'][0]
                ind = 288
                # read the number of stations
                station_number = np.frombuffer(
                    byteArray[ind:(ind+4)], dtype='i4')[0]
                ind += 4

                # read the number of elements

                element_number = np.frombuffer(
                    byteArray[ind:(ind+2)], dtype='i2')[0]

                if element_number == 0:
                    return None
                ind += 2

                # construct record structure
                element_type_map = {
                    1: 'b1', 2: 'i2', 3: 'i4', 4: 'i8', 5: 'f4', 6: 'f8', 7: 'S1'}
                element_map = {}
                element_map_len = {}
                for i in range(element_number):
                    element_id = str(np.frombuffer(byteArray[ind:(ind+2)], dtype='i2')[0])
                    ind += 2
                    element_type = np.frombuffer(
                        byteArray[ind:(ind+2)], dtype='i2')[0]
                    ind += 2
                    element_map[element_id] = element_type_map[element_type]
                    element_map_len[element_id] = int(element_type_map[element_type][1])

                if element_id0 is None:
                    dict0 = {}
                    id_dict = meteva.base.gds_element_id_dict
                    for key in element_map.keys():
                        if (int(key) in id_dict.values()):
                            for ele in id_dict.keys():
                                if int(key) == id_dict[ele]:
                                    dict0[ele] = key
                    if len(dict0.keys()) > 1:
                        print("element_id can not be None for this file")
                    else:
                        element_id_str0 = list(dict0.values())[0]
                dtype_str = element_map[element_id_str0]


                # loop every station to retrieve record
                if id_type ==0:
                    record_head_dtype = [
                        ('id', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
                    records = []
                    if station is None or len(station.index) * 100 > station_number:
                        for i in range(station_number):
                            record_head = np.frombuffer(
                                byteArray[ind:(ind+14)], dtype=record_head_dtype)
                            ind += 14
                            record = {
                                'id': record_head['id'][0], 'lon': record_head['lon'][0],
                                'lat': record_head['lat'][0]}
                            for j in range(record_head['numb'][0]):    # the record element number is not same, missing value is not included.
                                element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                                ind += 2
                                element_len = element_map_len[element_id]
                                if element_id == element_id_str0:
                                    record[data_name] = np.frombuffer(
                                        byteArray[ind:(ind + element_len)],
                                        dtype=dtype_str)[0]
                                    records.append(record)
                                ind += element_len
                        records = pd.DataFrame(records)
                        records.set_index('id')
                        # get time

                        records['time'] = time
                        records['level'] = level
                        records['dtime'] = dtime
                        new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', data_name]
                        records = records.reindex(columns=new_columns)

                        if station is None:
                            return records
                        else:
                            sta = meteva.base.put_stadata_on_station(records, station)
                            return sta
                    else:
                        sta = copy.deepcopy(station)
                        byte_num = len(byteArray)
                        i4_num = (byte_num - ind -4) //4
                        ids = np.zeros((i4_num,4),dtype=np.int32)

                        ids[:, 0] = np.frombuffer(byteArray[ind:(ind + i4_num * 4)], dtype='i4')
                        ids[:, 1] = np.frombuffer(byteArray[(ind +1):(ind + 1 + i4_num * 4)], dtype='i4')
                        ids[:, 2] = np.frombuffer(byteArray[(ind + 2):(ind + 2 + i4_num * 4)], dtype='i4')
                        ids[:, 3] = np.frombuffer(byteArray[(ind + 3):(ind + 3 + i4_num * 4)], dtype='i4')
                        ids = ids.flatten()
                        station_ids = station["id"].values
                        dat = np.zeros(station_ids.size)

                        for k in range(dat.size):
                            id1 = station_ids[k]
                            indexs = np.where(ids == id1)
                            if len(indexs[0]) >=1:
                                for n in range(len(indexs)):
                                    ind1 =ind +  indexs[n][0]
                                    record_head = np.frombuffer(byteArray[ind1:(ind1 + 14)], dtype=record_head_dtype)
                                    if(record_head['lon'][0] >=-180 and record_head['lon'][0] <= 360 and
                                            record_head['lat'][0] >= -90 and record_head['lat'][0] <= 90 and record_head["numb"][0] < 1000):
                                        ind1 += 14
                                        for j in range(record_head['numb'][0]):  # the record element number is not same, missing value is not included.
                                            element_id = str(np.frombuffer(byteArray[ind1:(ind1 + 2)], dtype='i2')[0])
                                            ind1 += 2
                                            element_len = element_map_len[element_id]
                                            if element_id == element_id_str0:
                                                sta.iloc[k,-1] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str)[0]
                                            ind1 += element_len
                        meteva.base.set_stadata_names(sta,[data_name])
                        sta['time'] = time
                        sta['level'] = level
                        sta['dtime'] = dtime
                        if show:
                            print("success read from " + filename)
                        return sta
                else:
                    record_head_dtype = [
                        ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
                    records = []
                    # if station is None or len(station.index) * 100 > station_number:
                    for i in range(station_number):
                        string_id_record_length = np.frombuffer(byteArray[ind:(ind + 2)], dtype="i2")[0]
                        dtype1 = "S" + str(string_id_record_length)
                        string_id = np.frombuffer(byteArray[ind + 2:(ind + 2 + string_id_record_length)], dtype=dtype1)[0]
                        string_id = string_id.decode()
                        ind += (2 + string_id_record_length)
                        record_head = np.frombuffer(
                            byteArray[ind:(ind + 10)], dtype=record_head_dtype)
                        ind += 10
                        record = {
                            'id': string_id, 'lon': record_head['lon'][0],
                            'lat': record_head['lat'][0]}

                        for j in range(record_head['numb'][
                                           0]):  # the record element number is not same, missing value is not included.
                            element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                            ind += 2
                            element_len = element_map_len[element_id]
                            if element_id == element_id_str0:
                                record[data_name] = np.frombuffer(
                                    byteArray[ind:(ind + element_len)],
                                    dtype=dtype_str)[0]
                                records.append(record)
                            ind += element_len
                    records = pd.DataFrame(records)
                    records.set_index('id')
                    # get time

                    records['time'] = time
                    records['level'] = level
                    records['dtime'] = dtime
                    new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', data_name]
                    records = records.reindex(columns=new_columns)
                    meteva.base.reset_id(records)
                    if station is None:
                        return records
                    else:
                        sta = meteva.base.put_stadata_on_station(records, station)
                        if show:
                            print("success read from " + filename)
                        return sta
            else:
                print("连接服务状态正常，但返回的输入内容为空")
                return None
        else:
            print("连接服务的状态异常，不能读取相应的文件,可能原因相应的文件不在允许读取的时段范围")
            return None
    except:
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        print(filename + "数据读取失败")
        return None




def read_stadata_from_gdsfile(filename,element_id = None,station = None, level=None,time=None, dtime=None, data_name='data0',show = False):

    element_id0 = element_id
    if element_id is not None:
        element_id_str0 = str(element_id)

    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        try:
            element_id_str0 = str(element_id)
            file = open(filename,"rb")
            byteArray = file.read()
            # define head structure
            head_dtype = [('discriminator', 'S4'), ('type', 'i2'),
                          ('description', 'S100'),
                          ('level', 'f4'), ('levelDescription', 'S50'),
                          ('year', 'i4'), ('month', 'i4'), ('day', 'i4'),
                          ('hour', 'i4'), ('minute', 'i4'), ('second', 'i4'),
                          ('Timezone', 'i4'),("id_type","i2"), ('extent', 'S98')]

            # read head information
            head_info = np.frombuffer(byteArray[0:288], dtype=head_dtype)
            if time is None:
                time = datetime.datetime(
                    head_info['year'][0], head_info['month'][0],
                    head_info['day'][0], head_info['hour'][0],
                    head_info['minute'][0], head_info['second'][0])
            else:
                time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
            if level is None:
                level = head_info["level"][0]
            if dtime is None:
                filename1 = os.path.split(filename)[1].split(".")
                dtime = int(filename1[1])
            id_type = head_info['id_type'][0]
            ind = 288
            # read the number of stations
            station_number = np.frombuffer(
                byteArray[ind:(ind+4)], dtype='i4')[0]
            ind += 4

            # read the number of elements
            element_number = np.frombuffer(
                byteArray[ind:(ind+2)], dtype='i2')[0]

            if element_number == 0:
                return None
            ind += 2

            # construct record structure
            element_type_map = {
                1: 'b1', 2: 'i2', 3: 'i4', 4: 'i8', 5: 'f4', 6: 'f8', 7: 'S1'}
            element_map = {}
            element_map_len = {}
            for i in range(element_number):
                element_id = str(np.frombuffer(byteArray[ind:(ind+2)], dtype='i2')[0])
                ind += 2
                element_type = np.frombuffer(
                    byteArray[ind:(ind+2)], dtype='i2')[0]
                ind += 2
                element_map[element_id] = element_type_map[element_type]
                element_map_len[element_id] = int(element_type_map[element_type][1])

            if element_id0 is None:
                dict0 = {}
                id_dict = meteva.base.gds_element_id_dict
                for key in element_map.keys():
                    if (int(key) in id_dict.values()):
                        for ele in id_dict.keys():
                            if int(key) == id_dict[ele]:
                                dict0[ele] = key
                if len(dict0.keys())>1:
                    print("element_id can not be None for this file" )
                else:
                    element_id_str0 = list(dict0.values())[0]
            dtype_str = element_map[element_id_str0]

            # loop every station to retrieve record
            if id_type ==0:
                record_head_dtype = [
                    ('id', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
                records = []
                if station is None or len(station.index) * 100 > station_number:
                    for i in range(station_number):
                        record_head = np.frombuffer(
                            byteArray[ind:(ind+14)], dtype=record_head_dtype)
                        ind += 14
                        record = {
                            'id': record_head['id'][0], 'lon': record_head['lon'][0],
                            'lat': record_head['lat'][0]}
                        for j in range(record_head['numb'][0]):    # the record element number is not same, missing value is not included.
                            element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                            ind += 2
                            element_len = element_map_len[element_id]
                            if element_id == element_id_str0:
                                record[data_name] = np.frombuffer(
                                    byteArray[ind:(ind + element_len)],
                                    dtype=dtype_str)[0]
                                records.append(record)
                            ind += element_len
                    records = pd.DataFrame(records)
                    records.set_index('id')
                    # get time
                    records['time'] = time
                    records['level'] = level
                    records['dtime'] = dtime
                    new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', data_name]
                    records = records.reindex(columns=new_columns)
                    if station is None:
                        return records
                    else:
                        sta = meteva.base.put_stadata_on_station(records, station)
                        if show:
                            print("success read from " + filename)
                        return sta
                else:
                    sta = copy.deepcopy(station)
                    byte_num = len(byteArray)
                    i4_num = (byte_num - ind -4) //4
                    ids = np.zeros((i4_num,4),dtype=np.int32)

                    ids[:, 0] = np.frombuffer(byteArray[ind:(ind + i4_num * 4)], dtype='i4')
                    ids[:, 1] = np.frombuffer(byteArray[(ind +1):(ind + 1 + i4_num * 4)], dtype='i4')
                    ids[:, 2] = np.frombuffer(byteArray[(ind + 2):(ind + 2 + i4_num * 4)], dtype='i4')
                    ids[:, 3] = np.frombuffer(byteArray[(ind + 3):(ind + 3 + i4_num * 4)], dtype='i4')
                    ids = ids.flatten()
                    station_ids = station["id"].values
                    dat = np.zeros(station_ids.size)

                    for k in range(dat.size):
                        id1 = station_ids[k]
                        indexs = np.where(ids == id1)
                        if len(indexs[0]) >=1:
                            for n in range(len(indexs)):
                                ind1 =ind +  indexs[n][0]
                                record_head = np.frombuffer(byteArray[ind1:(ind1 + 14)], dtype=record_head_dtype)
                                if(record_head['lon'][0] >=-180 and record_head['lon'][0] <= 360 and
                                        record_head['lat'][0] >= -90 and record_head['lat'][0] <= 90 and record_head["numb"][0] < 1000):
                                    ind1 += 14
                                    for j in range(record_head['numb'][0]):  # the record element number is not same, missing value is not included.
                                        element_id = str(np.frombuffer(byteArray[ind1:(ind1 + 2)], dtype='i2')[0])
                                        ind1 += 2
                                        element_len = element_map_len[element_id]
                                        if element_id == element_id_str0:
                                            sta.iloc[k,-1] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str)[0]
                                        ind1 += element_len
                    meteva.base.set_stadata_names(sta,[data_name])
                    sta['time'] = time
                    sta['level'] = level
                    sta['dtime'] = dtime
                    if show:
                        print("success read from " + filename)
                    return sta

            else:
                record_head_dtype = [
                   ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
                records = []
                #if station is None or len(station.index) * 100 > station_number:
                for i in range(station_number):
                    string_id_record_length = np.frombuffer( byteArray[ind:(ind + 2)],dtype="i2")[0]
                    dtype1 = "S" + str(string_id_record_length)
                    #print(string_id_record_length)
                    if string_id_record_length==0:
                        string_id ="999999"
                    else:
                        string_id = np.frombuffer(byteArray[ind+2:(ind + 2+string_id_record_length)],dtype=dtype1)[0]
                        string_id = string_id.decode()
                    ind += (2+string_id_record_length)
                    record_head = np.frombuffer(
                        byteArray[ind:(ind + 10)], dtype=record_head_dtype)
                    ind += 10
                    record = {
                        'id': string_id, 'lon': record_head['lon'][0],
                        'lat': record_head['lat'][0]}

                    for j in range(record_head['numb'][
                                       0]):  # the record element number is not same, missing value is not included.
                        element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                        ind += 2
                        element_len = element_map_len[element_id]
                        if element_id == element_id_str0:
                            record[data_name] = np.frombuffer(
                                byteArray[ind:(ind + element_len)],
                                dtype=dtype_str)[0]
                            records.append(record)
                        ind += element_len
                records = pd.DataFrame(records)
                records.set_index('id')
                # get time

                records['time'] = time
                records['level'] = level
                records['dtime'] = dtime
                new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', data_name]
                records = records.reindex(columns=new_columns)
                meteva.base.reset_id(records)
                if station is None:
                    return records
                else:
                    sta = meteva.base.put_stadata_on_station(records, station)
                    if show:
                        print("success read from " + filename)
                    return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename + "数据读取失败")
            return None


def read_stawind_from_gds(filename,station = None, level=None,time=None, dtime=None,data_name = "",show = True):

    if meteva.base.gds_ip_port is None:
        print("请先使用set_config 配置gds的ip和port")
        return
    directory, filename = os.path.split(filename)
    ip,port = meteva.base.gds_ip_port
    service = GDSDataService(ip, port)

    try:
        directory = directory.replace("mdfs:///", "")
        directory = directory.replace("\\","/")
        #print(filename)
        status, response = service.getData(directory, filename)

    except ValueError:
        print('Can not retrieve data' + filename + ' from ' + directory)
        return None
    try:
        ByteArrayResult = DataBlock_pb2.ByteArrayResult()
        if status == 200:
            ByteArrayResult.ParseFromString(response)
            if ByteArrayResult is not None:
                byteArray = ByteArrayResult.byteArray

                head_dtype = [('discriminator', 'S4'), ('type', 'i2'),
                          ('description', 'S100'),
                          ('level', 'f4'), ('levelDescription', 'S50'),
                          ('year', 'i4'), ('month', 'i4'), ('day', 'i4'),
                          ('hour', 'i4'), ('minute', 'i4'), ('second', 'i4'),
                          ('Timezone', 'i4'), ('extent', 'S100')]
                if(len(byteArray)<300):
                    return None
                # read head information
                head_info = np.frombuffer(byteArray[0:288], dtype=head_dtype)
                if time is None:
                    time = datetime.datetime(
                        head_info['year'][0], head_info['month'][0],
                        head_info['day'][0], head_info['hour'][0],
                        head_info['minute'][0], head_info['second'][0])
                else:
                    time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
                if level is None:
                    level = head_info["level"][0]
                if dtime is None:
                    filename1 = os.path.split(filename)[1].split(".")
                    dtime = int(filename1[1])
                ind = 288
                # read the number of stations
                station_number = np.frombuffer(
                    byteArray[ind:(ind+4)], dtype='i4')[0]
                ind += 4

                # read the number of elements
                element_number = np.frombuffer(
                    byteArray[ind:(ind+2)], dtype='i2')[0]

                if element_number == 0:
                    return None
                ind += 2

                # construct record structure
                element_type_map = {
                    1: 'b1', 2: 'i2', 3: 'i4', 4: 'i8', 5: 'f4', 6: 'f8', 7: 'S1'}
                element_map = {}
                element_map_len = {}
                for i in range(element_number):
                    element_id = str(np.frombuffer(byteArray[ind:(ind+2)], dtype='i2')[0])
                    ind += 2
                    element_type = np.frombuffer(
                        byteArray[ind:(ind+2)], dtype='i2')[0]
                    ind += 2
                    element_map[element_id] = element_type_map[element_type]
                    element_map_len[element_id] = int(element_type_map[element_type][1])

                dict0 = {}
                id_dict = meteva.base.gds_element_id_dict



                speed_id = -1
                angle_id = -1
                for key in element_map.keys():
                    if (int(key) in id_dict.values()):
                        for ele in id_dict.keys():
                            if int(key) == id_dict[ele]:
                                dict0[ele] = key
                                if ele.find("风速")>0:
                                    speed_id = key
                                    #print(ele)
                                if ele.find("风向")>0:
                                    angle_id = key
                                    #print(ele)

                if speed_id == -1 or angle_id == -1:
                    print("the file doesn't contains wind")
                dtype_str_speed = element_map[speed_id]
                dtype_str_angle = element_map[angle_id]

                # loop every station to retrieve record
                record_head_dtype = [
                    ('id', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
                records = []
                if station is None or len(station.index) * 100 > station_number:
                    for i in range(station_number):
                        record_head = np.frombuffer(
                            byteArray[ind:(ind+14)], dtype=record_head_dtype)
                        ind += 14
                        record = {
                            'id': record_head['id'][0], 'lon': record_head['lon'][0],
                            'lat': record_head['lat'][0]}
                        for j in range(record_head['numb'][0]):    # the record element number is not same, missing value is not included.
                            element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                            ind += 2
                            element_len = element_map_len[element_id]
                            hadwind = False
                            if element_id == speed_id:
                                record["speed"] = np.frombuffer(
                                    byteArray[ind:(ind + element_len)],
                                    dtype=dtype_str_speed)[0]
                                hadwind = True
                            if element_id == angle_id:
                                record["angle"] = np.frombuffer(
                                    byteArray[ind:(ind + element_len)],
                                    dtype=dtype_str_angle)[0]
                                hadwind = True
                            if hadwind:
                                records.append(record)
                            ind += element_len
                    records = pd.DataFrame(records)
                    records.set_index('id')
                    # get time

                    records['time'] = time
                    records['level'] = level
                    records['dtime'] = dtime
                    new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', "speed"+data_name,"angle"+data_name]
                    records = records.reindex(columns=new_columns)
                    if station is None:
                        return records
                    else:
                        sta = meteva.base.put_stadata_on_station(records, station)
                        if show:
                            print("success read from " + filename)
                        return sta
                else:
                    sta = copy.deepcopy(station)
                    meteva.base.set_stadata_names(sta,["speed"+data_name])
                    sta["angle"+data_name] = meteva.base.IV
                    byte_num = len(byteArray)
                    i4_num = (byte_num - ind -4) //4
                    ids = np.zeros((i4_num,4),dtype=np.int32)

                    ids[:, 0] = np.frombuffer(byteArray[ind:(ind + i4_num * 4)], dtype='i4')
                    ids[:, 1] = np.frombuffer(byteArray[(ind +1):(ind + 1 + i4_num * 4)], dtype='i4')
                    ids[:, 2] = np.frombuffer(byteArray[(ind + 2):(ind + 2 + i4_num * 4)], dtype='i4')
                    ids[:, 3] = np.frombuffer(byteArray[(ind + 3):(ind + 3 + i4_num * 4)], dtype='i4')
                    ids = ids.flatten()
                    station_ids = station["id"].values
                    dat = np.zeros(station_ids.size)

                    for k in range(dat.size):
                        id1 = station_ids[k]
                        indexs = np.where(ids == id1)
                        if len(indexs[0]) >=1:
                            for n in range(len(indexs)):
                                ind1 =ind +  indexs[n][0]
                                record_head = np.frombuffer(byteArray[ind1:(ind1 + 14)], dtype=record_head_dtype)
                                if(record_head['lon'][0] >=-180 and record_head['lon'][0] <= 360 and
                                        record_head['lat'][0] >= -90 and record_head['lat'][0] <= 90 and record_head["numb"][0] < 1000):
                                    ind1 += 14
                                    for j in range(record_head['numb'][0]):  # the record element number is not same, missing value is not included.
                                        element_id = str(np.frombuffer(byteArray[ind1:(ind1 + 2)], dtype='i2')[0])
                                        ind1 += 2
                                        element_len = element_map_len[element_id]
                                        if element_id == speed_id:
                                            sta.iloc[k,-2] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str_speed)[0]
                                        if element_id == angle_id:
                                            sta.iloc[k,-1] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str_angle)[0]
                                        ind1 += element_len
                    sta['time'] = time
                    sta['level'] = level
                    sta['dtime'] = dtime

                    if show:
                        print("success read from " + filename)
                    return sta
            else:
                print(filename + " not exist")
                return None
        else:
            print("连接服务的状态异常，不能读取相应的文件,可能原因相应的文件不在允许读取的时段范围")
            return None
    except:
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        print(filename + "数据读取失败")
        return None

def read_stawind_from_gdsfile(filename,station = None, level=None,time=None, dtime=None,data_name = "",show = False):
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else :
        try:
            file = open(filename,"rb")
            byteArray = file.read()
            # define head structure
            head_dtype = [('discriminator', 'S4'), ('type', 'i2'),
                          ('description', 'S100'),
                          ('level', 'f4'), ('levelDescription', 'S50'),
                          ('year', 'i4'), ('month', 'i4'), ('day', 'i4'),
                          ('hour', 'i4'), ('minute', 'i4'), ('second', 'i4'),
                          ('Timezone', 'i4'), ('extent', 'S100')]

            # read head information
            head_info = np.frombuffer(byteArray[0:288], dtype=head_dtype)
            if time is None:
                time = datetime.datetime(
                    head_info['year'][0], head_info['month'][0],
                    head_info['day'][0], head_info['hour'][0],
                    head_info['minute'][0], head_info['second'][0])
            else:
                time = meteva.base.tool.time_tools.all_type_time_to_time64(time)

            if level is None:
                level = head_info["level"][0]
            if dtime is None:
                filename1 = os.path.split(filename)[1].split(".")
                dtime = int(filename1[1])
            ind = 288
            # read the number of stations
            station_number = np.frombuffer(
                byteArray[ind:(ind+4)], dtype='i4')[0]
            ind += 4

            # read the number of elements
            element_number = np.frombuffer(
                byteArray[ind:(ind+2)], dtype='i2')[0]

            if element_number == 0:
                return None
            ind += 2

            # construct record structure
            element_type_map = {
                1: 'b1', 2: 'i2', 3: 'i4', 4: 'i8', 5: 'f4', 6: 'f8', 7: 'S1'}
            element_map = {}
            element_map_len = {}
            for i in range(element_number):
                element_id = str(np.frombuffer(byteArray[ind:(ind+2)], dtype='i2')[0])
                ind += 2
                element_type = np.frombuffer(
                    byteArray[ind:(ind+2)], dtype='i2')[0]
                ind += 2
                element_map[element_id] = element_type_map[element_type]
                element_map_len[element_id] = int(element_type_map[element_type][1])

            dict0 = {}
            id_dict = meteva.base.gds_element_id_dict



            speed_id = -1
            angle_id = -1
            for key in element_map.keys():
                if (int(key) in id_dict.values()):
                    for ele in id_dict.keys():
                        if int(key) == id_dict[ele]:
                            dict0[ele] = key
                            if ele.find("风速")>0:
                                speed_id = key
                                print(ele)
                            if ele.find("风向")>0:
                                angle_id = key
                                print(ele)

            if speed_id == -1 or angle_id == -1:
                print("the file doesn't contains wind")
            dtype_str_speed = element_map[speed_id]
            dtype_str_angle = element_map[angle_id]

            # loop every station to retrieve record
            record_head_dtype = [
                ('id', 'i4'), ('lon', 'f4'), ('lat', 'f4'), ('numb', 'i2')]
            records = []
            speed_name = "speed"+data_name
            angle_name = "angle"+data_name
            if station is None or len(station.index) * 100 > station_number:
                for i in range(station_number):
                    record_head = np.frombuffer(
                        byteArray[ind:(ind+14)], dtype=record_head_dtype)
                    ind += 14
                    record = {
                        'id': record_head['id'][0], 'lon': record_head['lon'][0],
                        'lat': record_head['lat'][0]}
                    for j in range(record_head['numb'][0]):    # the record element number is not same, missing value is not included.
                        element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
                        ind += 2
                        element_len = element_map_len[element_id]
                        hadwind = False
                        if element_id == speed_id:
                            record[speed_name] = np.frombuffer(
                                byteArray[ind:(ind + element_len)],
                                dtype=dtype_str_speed)[0]
                            hadwind = True
                        if element_id == angle_id:
                            record[angle_name] = np.frombuffer(
                                byteArray[ind:(ind + element_len)],
                                dtype=dtype_str_angle)[0]
                            hadwind = True
                        if hadwind:
                            records.append(record)
                        ind += element_len
                records = pd.DataFrame(records)
                records.set_index('id')
                # get time

                records['time'] = time
                records['level'] = level
                records['dtime'] = dtime
                new_columns = ['level', 'time', 'dtime', 'id', 'lon', 'lat', "speed"+data_name,"angle"+data_name]
                records = records.reindex(columns=new_columns)
                if station is None:
                    return records
                else:
                    sta = meteva.base.put_stadata_on_station(records, station)
                    if show:
                        print("success read from " + filename)
                    return sta
            else:
                sta = copy.deepcopy(station)
                meteva.base.set_stadata_names(sta,["speed"+data_name])
                sta["angle"+data_name] = meteva.base.IV
                byte_num = len(byteArray)
                i4_num = (byte_num - ind -4) //4
                ids = np.zeros((i4_num,4),dtype=np.int32)

                ids[:, 0] = np.frombuffer(byteArray[ind:(ind + i4_num * 4)], dtype='i4')
                ids[:, 1] = np.frombuffer(byteArray[(ind +1):(ind + 1 + i4_num * 4)], dtype='i4')
                ids[:, 2] = np.frombuffer(byteArray[(ind + 2):(ind + 2 + i4_num * 4)], dtype='i4')
                ids[:, 3] = np.frombuffer(byteArray[(ind + 3):(ind + 3 + i4_num * 4)], dtype='i4')
                ids = ids.flatten()
                station_ids = station["id"].values
                dat = np.zeros(station_ids.size)

                for k in range(dat.size):
                    id1 = station_ids[k]
                    indexs = np.where(ids == id1)
                    if len(indexs[0]) >=1:
                        for n in range(len(indexs)):
                            ind1 =ind +  indexs[n][0]
                            record_head = np.frombuffer(byteArray[ind1:(ind1 + 14)], dtype=record_head_dtype)
                            if(record_head['lon'][0] >=-180 and record_head['lon'][0] <= 360 and
                                    record_head['lat'][0] >= -90 and record_head['lat'][0] <= 90 and record_head["numb"][0] < 1000):
                                ind1 += 14
                                for j in range(record_head['numb'][0]):  # the record element number is not same, missing value is not included.
                                    element_id = str(np.frombuffer(byteArray[ind1:(ind1 + 2)], dtype='i2')[0])
                                    ind1 += 2
                                    element_len = element_map_len[element_id]
                                    if element_id == speed_id:
                                        sta.iloc[k,-2] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str_speed)[0]
                                    if element_id == angle_id:
                                        sta.iloc[k,-1] = np.frombuffer(byteArray[ind1:(ind1 + element_len)],dtype=dtype_str_angle)[0]
                                    ind1 += element_len
                sta['time'] = time
                sta['level'] = level
                sta['dtime'] = dtime
                if show:
                    print("success read from " + filename)
                return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename + "数据读取失败")
            return None


def read_stadata_from_gds_griddata(filename,station,level = None,time =None,dtime = None,data_name = "data0",show = False):
    # ip 为字符串形式，示例 “10.20.30.40”
    # port 为整数形式
    # filename 为字符串形式 示例 "ECMWF_HR/TCDC/19083108.000"
    if meteva.base.gds_ip_port is None:
        print("请先使用set_config 配置gds的ip和port")
        return
    ip,port = meteva.base.gds_ip_port
    service = GDSDataService(ip, port)
    try:
        if(service is None):
            print("service is None")
            return
        filename = filename.replace("mdfs:///", "")
        filename = filename.replace("\\","/")
        directory,fileName = os.path.split(filename)
        status, response = service.getData(directory, fileName)
        ByteArrayResult = DataBlock_pb2.ByteArrayResult()
        if status == 200:
            ByteArrayResult.ParseFromString(response)
            if ByteArrayResult is not None:
                byteArray = ByteArrayResult.byteArray
                level1, y, m, d, h, timezone, period = struct.unpack("fiiiiii", byteArray[106:134])
                startLon, endLon, dlon, nlon= struct.unpack("fffi", byteArray[134:150])
                startLat, endLat, dlat, nlat = struct.unpack("fffi", byteArray[150:166])
                nsta = len(station.index)
                ig = ((station['lon'].values - startLon) // dlon).astype(dtype='int32')
                jg = ((station['lat'].values - startLat) // dlat).astype(dtype='int32')
                dx = (station['lon'].values - startLon) / dlon - ig
                dy = (station['lat'].values - startLat) / dlat - jg
                c00 = (1 - dx) * (1 - dy)
                c01 = dx * (1 - dy)
                c10 = (1 - dx) * dy
                c11 = dx * dy
                ig1 = np.minimum(ig + 1, nlon - 1)
                jg1 = np.minimum(jg + 1, nlat - 1)
                i00 = (nlon * jg + ig)
                i01 = nlon * jg+ ig1
                i10 = nlon * jg1 + ig
                i11 = nlon * jg1 + ig1
                dat = np.zeros(nsta)
                #i4 = np.arange(4)
                #xx,yy = np.meshgrid(i4,i00)
                #i00 = xx + yy
                #i00 = i00.flatten()
                for i in range(nsta):
                    dat00 = np.frombuffer(byteArray[278 + i00[i] * 4:i00[i] * 4 + 282], dtype='float32')
                    dat01 = np.frombuffer(byteArray[278 + i01[i] * 4:i01[i] * 4 + 282], dtype='float32')
                    dat10 = np.frombuffer(byteArray[278 + i10[i] * 4:i10[i] * 4 + 282], dtype='float32')
                    dat11 = np.frombuffer(byteArray[278 + i11[i] * 4:i11[i] * 4 + 282], dtype='float32')
                    dat[i] = c00[i] * dat00 + c01[i] * dat01 +c10[i] * dat10 + c11[i] * dat11
                #grd.values = np.frombuffer(byteArray[278:], dtype='float32')
                sta = copy.deepcopy(station)
                sta.iloc[:,-1] = dat[:]
                filename1 = os.path.split(filename)[1].split(".")
                #print(filename1)
                if time is None:
                    time = datetime.datetime(y, m, d, h, 0)
                else:
                    time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
                if level is None:
                    level = level1
                if dtime is None:
                    dtime = period
                sta.loc[:, "level"] = level
                sta.loc[:, "time"] = time
                sta.loc[:, "dtime"] = dtime
                meteva.base.set_stadata_names(sta,[data_name])
                if show:
                    print("success read from " + filename)
                return sta
            else:
                print(filename + " not exist")
                return None
        else:
            print("连接服务的状态异常，不能读取相应的文件,可能原因相应的文件不在允许读取的时段范围")
            return None
    except :
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        print(filename +" 数据读取失败")
        return None

def print_gds_file_values_names(filename):
    # ip 为字符串形式，示例 “10.20.30.40”
    # port 为整数形式
    # filename 为字符串形式 示例 "ECMWF_HR/TCDC/19083108.000"
    value_id_list= []
    if os.path.exists(filename):
        file = open(filename, "rb")
        byteArray = file.read()
    else:

    #if os.path.exists(filename):
    #    ip,port = None,None
    #else:
        print(filename + "不是本地文件,尝试从服务器上读取相应数据")
        if meteva.base.gds_ip_port is None:
            print("在本地找不到文件"+filename+"考虑它为服务器上的路径")
            print("请先使用set_config 配置gds的ip和port")
            return
        ip,port = meteva.base.gds_ip_port

        filename = filename.replace("mdfs:///", "")
        filename = filename.replace("\\", "/")

        if ip is not None:
            service = GDSDataService(ip, port)
            try:
                directory, fileName = os.path.split(filename)
                status, response = service.getData(directory, fileName)
                ByteArrayResult = DataBlock_pb2.ByteArrayResult()
                if status == 200:
                    ByteArrayResult.ParseFromString(response)
                    if ByteArrayResult is not None:
                        byteArray = ByteArrayResult.byteArray
                else:
                    print("数据内容不可读")
            except:
                exstr = traceback.format_exc()
                print(exstr)


    ind = 288
    # read the number of stations
    station_number = np.frombuffer(
        byteArray[ind:(ind + 4)], dtype='i4')[0]
    ind += 4

    # read the number of elements
    element_number = np.frombuffer(
        byteArray[ind:(ind + 2)], dtype='i2')[0]

    if element_number == 0:
        return None
    ind += 2

    # construct record structure
    element_type_map = {
        1: 'b1', 2: 'i2', 3: 'i4', 4: 'i8', 5: 'f4', 6: 'f8', 7: 'S1'}
    element_map = {}
    element_map_len = {}
    for i in range(element_number):
        element_id = str(np.frombuffer(byteArray[ind:(ind + 2)], dtype='i2')[0])
        ind += 2
        element_type = np.frombuffer(
            byteArray[ind:(ind + 2)], dtype='i2')[0]
        ind += 2
        element_map[element_id] = element_type_map[element_type]
        element_map_len[element_id] = int(element_type_map[element_type][1])
    id_dict = meteva.base.gds_element_id_dict

    dict0 = {}
    for key in element_map.keys():
        if(int(key) in id_dict.values()):
            for ele in id_dict.keys():
                if int(key) == id_dict[ele]:
                    dict0[ele] = int(key)
                    print(ele + ":" + key)
    return dict0

def read_stadata_from_micaps16(filename,level = None,time= None,dtime = None,data_name = "data0",show = False):
    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        encoding,_ = meteva.base.io.get_encoding_of_file(filename,read_rows=1)
        if encoding is None:
            print("文件编码格式不识别")
            return None
        try:
            file = open(filename, 'r',encoding=encoding)
            head = file.readline()
            head = file.readline()
            stationids = []
            row1 = []
            row2 = []
            row3 = []
            while(head is not None and head.strip() != ""):
                strs = head.split()
                stationids.append(strs[0])
                a = int(strs[1])
                b = a // 100 + (a % 100) /60
                row1.append(b)
                a = int(strs[2])
                b = a // 100 + (a % 100) /60
                row2.append(b)
                row3.append(float(strs[3]))
                head =  file.readline()

            row1 = np.array(row1)
            row2 = np.array(row2)
            row3 = np.array(row3)
            ids = np.array(stationids)
            dat = np.zeros((len(row1),4))
            dat[:,0] = ids[:]
            if(np.max(row2) > 90 or np.min(row2) <-90):
                dat[:,1] = row2[:]
                dat[:,2] = row1[:]
            else:
                dat[:,1] = row1[:]
                dat[:,2] = row2[:]
            dat[:,3] = row3[:]
            station = pd.DataFrame(dat, columns=['id','lon', 'lat', data_name])
            station = meteva.base.sta_data(station)
            meteva.base.set_stadata_coords(station,level=level,time= time,dtime = dtime)
            if show:
                print("success read from " + filename)
            return station
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename+"文件格式不能识别。可能原因：文件未按micaps16格式存储")
            return None


def read_stadata_from_gds_griddata_file(filename,station,level = None,time = None,dtime = None,data_name = "data0",show = False):

    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None

    else:
        try:
            file = open(filename,"rb")
            position = file.seek(106)
            content = file.read(28)
            level1, y, m, d, h, timezone, period = struct.unpack("fiiiiii", content)
            position = file.seek(134)
            content = file.read(32)
            slon, elon, dlon,nlon,slat,elat,dlat,nlat = struct.unpack("fffifffi", content)
            nsta = len(station.index)
            ig = ((station['lon'].values - slon) // dlon).astype(dtype='int32')
            jg = ((station['lat'].values - slat) // dlat).astype(dtype='int32')
            dx = (station['lon'].values - slon) / dlon - ig
            dy = (station['lat'].values - slat) / dlat - jg
            c00 = (1 - dx) * (1 - dy)
            c01 = dx * (1 - dy)
            c10 = (1 - dx) * dy
            c11 = dx * dy
            ig1 = np.minimum(ig + 1, nlon - 1)
            jg1 = np.minimum(jg + 1, nlat - 1)
            i00 = (nlon * jg + ig)
            i01 = nlon * jg + ig1
            i10 = nlon * jg1 + ig
            i11 = nlon * jg1 + ig1
            dat = np.zeros(nsta)
            for i in range(nsta):
                position = file.seek(i00[i] * 4+278)
                content = file.read(4)
                dat00 = np.frombuffer(content, dtype='float32')
                position = file.seek(i01[i] * 4+278)
                content = file.read(4)
                dat01 = np.frombuffer(content, dtype='float32')
                position = file.seek(i10[i] * 4+278)
                content = file.read(4)
                dat10 = np.frombuffer(content, dtype='float32')
                position = file.seek(i11[i] * 4+278)
                content = file.read(4)
                dat11 = np.frombuffer(content, dtype='float32')
                dat[i] = c00[i] * dat00 + c01[i] * dat01 + c10[i] * dat10 + c11[i] * dat11
            file.close()
            sta = copy.deepcopy(station)
            sta.iloc[:, -1] = dat[:]
            if time is None:
                time = datetime.datetime(y, m, d, h, 0)
            else:
                time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
            if level is None:
                level = level1
            if dtime is None:
                dtime = period
            sta.loc[:, "level"] = level
            sta.loc[:, "time"] = time
            sta.loc[:, "dtime"] = dtime
            meteva.base.set_stadata_names(sta, [data_name])

            if show:
                print("success read from " + filename)
            return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename+"文件格式不能识别")
            return None


def read_stawind_from_gds_gridwind_file(filename,station,level = None,time = None,dtime = None,data_name = "",show = False):

    if not os.path.exists(filename):
        print(filename+"文件不存在")
        return None
    elif os.path.isdir(filename):
        print(filename+"是文件夹而不是文件")
        return None
    else:
        try:
            file = open(filename, "rb")
            position = file.seek(106)
            content = file.read(28)
            level1, y, m, d, h, timezone, period = struct.unpack("fiiiiii", content)
            position = file.seek(134)
            content = file.read(32)
            slon, elon, dlon, nlon, slat, elat, dlat, nlat = struct.unpack("fffifffi", content)
            nsta = len(station.index)
            ig = ((station['lon'].values - slon) // dlon).astype(dtype='int32')
            jg = ((station['lat'].values - slat) // dlat).astype(dtype='int32')
            dx = (station['lon'].values - slon) / dlon - ig
            dy = (station['lat'].values - slat) / dlat - jg
            c00 = (1 - dx) * (1 - dy)
            c01 = dx * (1 - dy)
            c10 = (1 - dx) * dy
            c11 = dx * dy
            ig1 = np.minimum(ig + 1, nlon - 1)
            jg1 = np.minimum(jg + 1, nlat - 1)
            i00 = (nlon * jg + ig)
            i01 = nlon * jg + ig1
            i10 = nlon * jg1 + ig
            i11 = nlon * jg1 + ig1
            udata = np.zeros(nsta)
            vdata = np.zeros(nsta)
            i_s0 = 278
            i_s1 = i_s0 + nlon * nlat * 4

            for i in range(nsta):
                position = file.seek(i00[i] * 4 + i_s0)
                content = file.read(4)
                dat00 = np.frombuffer(content, dtype='float32')
                position = file.seek(i01[i] * 4 + i_s0)
                content = file.read(4)
                dat01 = np.frombuffer(content, dtype='float32')
                position = file.seek(i10[i] * 4 + i_s0)
                content = file.read(4)
                dat10 = np.frombuffer(content, dtype='float32')
                position = file.seek(i11[i] * 4 + i_s0)
                content = file.read(4)
                dat11 = np.frombuffer(content, dtype='float32')
                udata[i] = c00[i] * dat00 + c01[i] * dat01 + c10[i] * dat10 + c11[i] * dat11

                position = file.seek(i00[i] * 4 + i_s1)
                content = file.read(4)
                dat00 = np.frombuffer(content, dtype='float32')
                position = file.seek(i01[i] * 4 + i_s1)
                content = file.read(4)
                dat01 = np.frombuffer(content, dtype='float32')
                position = file.seek(i10[i] * 4 + i_s1)
                content = file.read(4)
                dat10 = np.frombuffer(content, dtype='float32')
                position = file.seek(i11[i] * 4 + i_s1)
                content = file.read(4)
                dat11 = np.frombuffer(content, dtype='float32')
                vdata[i] = c00[i] * dat00 + c01[i] * dat01 + c10[i] * dat10 + c11[i] * dat11
            file.close()
            sta = copy.deepcopy(station)
            sta.iloc[:, -1] = udata[:]
            sta["v"] = vdata
            if time is None:
                time = datetime.datetime(y, m, d, h, 0)
            else:
                time = meteva.base.tool.time_tools.all_type_time_to_time64(time)
            if level is None:
                level = level1
            if dtime is None:
                dtime = period
            sta.loc[:, "level"] = level
            sta.loc[:, "time"] = time
            sta.loc[:, "dtime"] = dtime
            meteva.base.set_stadata_names(sta,["u"+data_name,"v"+data_name])

            if show:
                print("success read from " + filename)
            return sta
        except:
            if show:
                exstr = traceback.format_exc()
                print(exstr)
            print(filename + "文件格式不能识别")
            return None


def read_stadata_from_cmadaas(dataCode,element,time,station = None,level=0,dtime=0,data_name= None,show = False,id_type = "Station_Id_D"):
    ## stations
    qparams = {'interfaceId': 'getSurfEleByTime',
               'dataCode': dataCode,
               'elements': 'Datetime,' + id_type+',Lat,Lon,'+element,
               }  ##字典规定接口名称，数据代码，下载要素代码。
    # 数据部分： SURF_CHN_MUL_HOR_N 数据为全国基准站(2400多站点)逐小时地面要素
    # 接口部分： getSurfEleByTime 接口为按时间提取地面要素
    # 要素部分： PRE_24h 为地面降水，其他包括TEM、TEM_Max、RHU、WIN_D_Avg_10mi、WIN_S_Avg_10mi等，可自己选择
    time = meteva.base.all_type_time_to_datetime(time)
    time_str = time.strftime("%Y%m%d%H%M")
    userID = meteva.base.cmadaas_set[2]
    pwd = meteva.base.cmadaas_set[3]
    try:
        sta = CMADaasAccess.get_obs_micaps3_from_cmadaas(qparams, userId=userID, pwd=pwd, time=time_str,show=show)
    except:
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        return None
    if sta is not None:
        if data_name is None:
            data_name = dataCode
        meteva.base.set_stadata_names(sta,data_name_list=[data_name])
        sta['level'] = level
        sta['dtime'] = dtime
        if (station is not None):
            sta = meteva.base.put_stadata_on_station(sta, station)
    else:
        print("数据读取失败")
    return sta


def read_stadata_from_cimiss(dataCode,element,time,station = None,level = 0,dtime = 0,show = False):

    time1 = meteva.base.all_type_time_to_datetime(time)
    time_str = time1.strftime("%Y%m%d%H%M%S")
    if dataCode.find("UPAR_")>=0:
        interface_id = "getUparEleByTime"
    else:
        interface_id = "getSurfEleByTime"
    params = {'dataCode': dataCode,
              'times': time_str,
              'orderby': "Station_Id_d",
              'elements': "Station_Id_d,lon,lat,"+element}

    contents = get_http_result_cimiss(interface_id,params,show_url=show)
    if contents is None:
        return None
    try:
        contents = json.loads(contents.decode('utf-8'))
        if contents['returnCode'] != '0':
            return None

        # construct pandas DataFrame
        df = pd.DataFrame(contents['DS'])
        data1 =df.iloc[0,3]
        if isinstance(data1,str):
            df[element] = df[element].astype("float")
        sta = meteva.base.sta_data(df,columns=["id","lon","lat",element])

        #sta.time = pd.to_datetime(sta.time)
        #sta.level = sta.level.astype(np.int16)
        sta.id = sta.id.astype(np.int32)
        #sta.dtime = sta.dtime.astype(np.int16)
        sta.lon = sta.lon.astype(np.float32)
        sta.lat = sta.lat.astype(np.float32)
        #sta.data0 = sta.data0.astype(np.float32)
        #sta.data0[sta.data0 >= 10000] = default

        meteva.base.set_stadata_coords(sta,time = time1,dtime=dtime,level=level)
        if (station is not None):
            sta = meteva.base.put_stadata_on_station(sta, station)
        return sta
    except:
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        return None


def read_cyclone_trace(filename, id_cyclone,column=8,  data_name="data0",show = False):
    try:
        column_all = [0, 1, 2, 3, 4, 5, 6]
        if isinstance(column, list):
            column_all.extend(column)
        else:
            column_all.append(column)

        sta1 = pd.read_csv(filename, skiprows=2, sep="\s+", header=None,
                           usecols=column_all)

        dict1 = meteva.base.m7_element_column_dict
        dict2 = dict([val, key] for key, val in dict1.items())

        if len(column_all) == 8:
            dat_dict = {"level": {}, "time": {}, "dtime": {}, "id": {}, "lon": {}, "lat": {}, data_name: {} }
        else:
            dat_dict = {"level": {}, "time": {}, "dtime": {}, "id": {}, "lon": {}, "lat": {}}
            for j in range(7, len(column_all)):
                k = column_all[j]
                name = dict2[k]
                dat_dict[name] = {}

        for i in range(len(sta1.index)):
            dat_dict["level"][i] = 0
            dat_dict["time"][i] = datetime.datetime(sta1.iloc[i, 0], sta1.iloc[i, 1], sta1.iloc[i, 2], sta1.iloc[i, 3])
            dat_dict["dtime"][i] = sta1.iloc[i, 4]
            dat_dict["id"][i] = id_cyclone
            dat_dict["lon"][i] = sta1.iloc[i, 5]
            dat_dict["lat"][i] = sta1.iloc[i, 6]
            if len(column_all) == 8:
                dat_dict[data_name][i] = sta1.iloc[i, -1]
            else:
                for j in range(7,len(column_all)):
                    k = column_all[j]
                    name = dict2[k]
                    dat_dict[name][i] = sta1.iloc[i, j]

        sta2 = pd.DataFrame(dat_dict)
        # print(sta2)
        return sta2
    except:
        if show:
            exstr = traceback.format_exc()
            print(exstr)
        return None


import math


def as_psp(x):
    result = {}
    angle = x['angle']
    x1 = (math.cos(angle)) * x['length'] / 2 + x['xmid']
    x0 = x['xmid'] - math.cos(angle) * x['length'] / 2
    y0 = - math.sin(angle) * x['length'] / 2 + x['ymid']
    y1 = x['ymid'] + math.sin(angle) * x['length'] / 2
    result['ends'] = {'x0': x0, 'y0': y0, 'x1': x1, 'y1': y1}
    return result



import meteva
import numpy as np
from scipy.ndimage.filters import uniform_filter
import datetime
import pandas as pd

def fss_merge(fss_df,s = None):
    pass


def fss(grd_ob,grd_fo,grade_list=[1e-30],half_window_size_list=[1],compare = ">=", masker=None):
    '''
    :param grd_ob:
    :param grd_fo:
    :param grade_list:
    :param half_window_sizes_list:
    :param compare:
    :param masker_xy:
    :return:
    '''

    grd_ob_list = meteva.base.fun.split_grd(grd_ob,used_coords = ["time"])
    grd_ob_dict = {}
    for i in range(len(grd_ob_list)):
        grd_ob_one = grd_ob_list[i]
        time1 = meteva.base.all_type_time_to_datetime(grd_ob_one["time"].values[0])
        grd_ob_dict[time1] = grd_ob_one
    grd_fo_list = meteva.base.fun.split_grd(grd_fo)
    grid0 = meteva.base.get_grid_of_data(grd_fo_list[0])
    grid0 = meteva.base.grid(grid0.glon,grid0.glat)

    if masker is not None:
        masker1 = meteva.base.interp_gs_nearest(masker,grid0)
        masker_xy = masker1.values.squeeze()
    else:
        masker_xy = masker


    nw = len(half_window_size_list)
    nt = len(grade_list)
    result = []
    for i in range(len(grd_fo_list)):
        grd_fo_one = grd_fo_list[i]
        time1 = grd_fo_one["time"].values[0]
        dtime1 = int(grd_fo_one["dtime"].values[0])
        member1 = grd_fo_one["member"].values[0]
        time_ob = meteva.base.all_type_time_to_datetime(time1) + datetime.timedelta(hours=dtime1)
        if time_ob in grd_ob_dict.keys():
            grd_ob_one = grd_ob_dict[time_ob]
            ob_xy = grd_ob_one.values.squeeze()
            fo_xy = grd_fo_one.values.squeeze()
            fbs_pobfo_array = fbs_pobfo(ob_xy,fo_xy,grade_list = grade_list,half_window_size_list = half_window_size_list,
                                              compare = compare,masker = masker_xy)

            fss1 = 1 - fbs_pobfo_array[...,2]/(fbs_pobfo_array[...,0]+fbs_pobfo_array[...,1] + 1e-30)
            for j in range(nw):
                for k in range(nt):
                    result_one = {"time":time1,"dtime":dtime1,"fname":member1,
                                  "half_window_size":half_window_size_list[j],"grade":grade_list[k],
                                  "pob":fbs_pobfo_array[j,k,0],
                                  "pfo":fbs_pobfo_array[j,k,1],
                                  "fbs": fbs_pobfo_array[j, k, 2],
                                  "fss":fss1[j,k]}
                    result.append(result_one)
    df = pd.DataFrame(result)
    result_sta = meteva.base.sta_data(df)
    return result_sta




def fbs_pobfo(ob_xy, fo_xy,grade_list=[1e-30],half_window_size_list=[1],compare = ">=", masker=None):
    '''
    :param Ob: 实况数据 2维的numpy
    :param Fo: 实况数据 2维的numpy
    :param window_sizes_list: 卷积窗口宽度的列表，以格点数为单位
    :param threshold_list:  事件发生的阈值
    :param Masker:  2维的numpy检验的关注区域，在Masker网格值取值为0或1，函数只对网格值等于1的区域的数据进行计算。
    :return:
    '''
    def moving_ave(dat_xy, half_window_size):
        size = half_window_size * 2 + 1
        dat1 = uniform_filter(dat_xy, size=size)
        dat1 = np.round(dat1[:, :], 10)
        return dat1

    if compare not in [">=",">","<","<="]:
        print("compare 参数只能是 >=   >  <  <=  中的一种")
        return
    shape = ob_xy.shape
    nw = len(half_window_size_list)
    nt = len(grade_list)
    result = np.zeros((nw,nt,3))
    if masker is None:
        count = ob_xy.size
    else:
        count = np.sum(masker)

    for j in range(nt):
        ob_01 = np.zeros(shape)
        fo_01 = np.zeros(shape)
        if compare == ">=":
            ob_01[ob_xy>=grade_list[j]]  = 1
            fo_01[fo_xy>=grade_list[j]]  = 1
        elif compare =="<=":
            ob_01[ob_xy<=grade_list[j]]  = 1
            fo_01[fo_xy<=grade_list[j]]  = 1
        elif compare ==">":
            ob_01[ob_xy>grade_list[j]]  = 1
            fo_01[fo_xy>grade_list[j]]  = 1
        else:
            ob_01[ob_xy<grade_list[j]]  = 1
            fo_01[fo_xy<grade_list[j]]  = 1
        for i in range(nw):
            ob_01_smooth = moving_ave(ob_01, half_window_size_list[i])
            fo_01_smooth = moving_ave(fo_01, half_window_size_list[i])
            if masker is not None:
                ob_01_smooth *= masker
                fo_01_smooth *= masker
            result[i,j,2] = np.sum(np.square(ob_01_smooth - fo_01_smooth))
            result[i,j,0]  = np.sum(np.square(ob_01_smooth))
            result[i,j,1]  = np.sum(np.square(fo_01_smooth))

    result /= count
    return result




def fss_time_base_on_mid(mid_array):
    '''

    :param mid_array:
    :return:
    '''
    result = 1 - mid_array[..., 0] / (mid_array[..., 1] + 1e-30)
    return result



def fss_time(Ob,Fo,grade_list = [1e-30],compare =">-",window_size = None):
    '''
    :param Ob: 二维numpy数组Ob[i,j]，其中i取值为0  - 站点数， j为取值为0 - 时效维度的size
    :param Fo: 二维numpy数组Fo[i,j]，其中i取值为0  - 站点数， j为取值为0 - 时效维度的size
    :param window_size:
    :param grade_list:
    :return:
    '''
    mid_array = mid_fss_time(Ob,Fo,grade_list,window_size=window_size,compair = compare)
    result = fss_time_base_on_mid(mid_array)
    return result

#

def merge_mid_fss_time(mid_array1,mid_array2):
    '''

    :param mid_array1:
    :param mid_array2:
    :return:
    '''
    if mid_array1 is None:
        return mid_array2
    if mid_array2 is None:
        return mid_array1
    return mid_array1 + mid_array2

def mid_fss_time(Ob,Fo,grade_list = [1e-30],compare =">-",compair = ">=",window_size = None):
    '''
    :param Ob: 二维numpy数组Ob[i,j]，其中i取值为0  - 站点数， j为取值为0 - 时效维度的size
    :param Fo: 二维numpy数组Fo[i,j]，其中i取值为0  - 站点数， j为取值为0 - 时效维度的size
    :param window_size:
    :param grade_list:
    :return:
    '''
    if compair not in [">=",">","<","<="]:
        print("compair 参数只能是 >=   >  <  <=  中的一种")
        return
    shape = Ob.shape
    if len(shape) == 1:
        Ob = Ob.reshape(1,shape[0])
        Fo = Fo.reshape(1,shape[0])
        shape = Ob.shape
    if window_size is None:
        window_size = shape[1]//3
    left_size = shape[1] - window_size
    ng = len(grade_list)
    result = np.zeros((ng,window_size,left_size,2))
    for i in range(1,1+window_size):
        for j in range(left_size):
            for g in range(ng):
                ob1 = np.zeros((shape[0],i))
                fo1 = np.zeros((shape[0],i))
                if compair == ">=":
                    ob1[Ob[:, j + window_size -i:j+window_size] >= grade_list[g]] = 1
                    fo1[Fo[:, j + window_size -i:j+window_size] >= grade_list[g]] = 1
                elif compair == "<=":
                    ob1[Ob[:, j + window_size -i:j+window_size] <= grade_list[g]] = 1
                    fo1[Fo[:, j + window_size -i:j+window_size] <= grade_list[g]] = 1
                elif compair == ">":
                    ob1[Ob[:, j + window_size -i:j+window_size] > grade_list[g]] = 1
                    fo1[Fo[:, j + window_size -i:j+window_size] > grade_list[g]] = 1
                elif compair == "<":
                    ob1[Ob[:, j + window_size -i:j+window_size] < grade_list[g]] = 1
                    fo1[Fo[:, j + window_size -i:j+window_size] < grade_list[g]] = 1

                ob_hap_p =np.sum(ob1,axis=1)
                fo_hap_p =np.sum(fo1,axis=1)
                result[g,i - 1, j,  0] = np.sum(np.power(ob_hap_p - fo_hap_p, 2))
                result[g,i - 1, j,  1] = np.sum(np.power(ob_hap_p, 2)) + np.sum(np.power(fo_hap_p, 2))
    return result


from . import score
from . import table
from . import plot
from .score import tems,tems_merge,bs,bs_tems,bss_tems,bss,roc_auc_hnh,roc_auc
from .table import *
from .plot import reliability,discrimination,reliability_hnh,discrimination_hnh,roc_hfmc,roc_hnh,comprehensive_hnh,comprehensive_probability,roc


from .transformating import *
from .interpolating import *
from .combining import *
from . import  selecting  as sele
from . import computing  as comp
from . import diagnosing as diag
from . import statisticing as stat
from . import nearing as near
from . import timing as timing
from .selecting import *
from .computing import *
from .diagnosing import *
from . import grouping as grp
from .grouping import group,split,group_grd,split_grd
from .statisticing import *
from .nearing import *
from .timing import *




import meteva
from meteva.base.tool.math_tools import lon_lat_to_cartesian
from scipy.spatial import cKDTree
import copy
import numpy as np
import pandas as pd

def sta_index_ensemble_near_by_sta(sta_to,nearNum = 100,sta_from = None,drop_frist = False):
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    _,indexs = tree.query(xyz_sta0, k=nearNum)
    sta_ensemble = sta_to[meteva.base.get_coord_names()]
    for i in range(nearNum):
        data_name = "data" + str(i)
        if nearNum ==1:
            sta_ensemble[data_name] = indexs[:]
        else:
            sta_ensemble[data_name] = indexs[:, i]
    if drop_frist:
        sta_ensemble = sta_ensemble.drop(columns=['data0'])
    return sta_ensemble

def sta_id_ensemble_near_by_sta(sta_to,nearNum = 100,sta_from = None,drop_frist = False):
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    _,indexs = tree.query(xyz_sta0, k=nearNum)
    input_dat = sta_from.loc[:, 'id'].values
    sta_ensemble = sta_to[meteva.base.get_coord_names()]
    for i in range(nearNum):
        data_name = "data" + str(i)
        if nearNum ==1:
            sta_ensemble[data_name] = input_dat[indexs[:]]
        else:
            sta_ensemble[data_name] = input_dat[indexs[:, i]]
    if drop_frist:
        sta_ensemble = sta_ensemble.drop(columns=['data0'])
    return sta_ensemble

def sta_value_ensemble_near_by_sta(sta_to,nearNum = 100,sta_from = None,drop_frist = False):
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    _,indexs = tree.query(xyz_sta0, k=nearNum)
    data_name = meteva.base.get_stadata_names(sta_from)[0]
    input_dat = sta_from[data_name].values
    sta_ensemble = sta_to[meteva.base.get_coord_names()]
    for i in range(nearNum):
        data_name = "data" + str(i)
        if nearNum ==1:
            sta_ensemble[data_name] = input_dat[indexs[:]]
        else:
            sta_ensemble[data_name] = input_dat[indexs[:, i]]
    if drop_frist:
        sta_ensemble = sta_ensemble.drop(columns=['data0'])
    return sta_ensemble

def sta_dis_ensemble_near_by_sta(sta_to,nearNum = 100,sta_from = None,drop_frist = False):
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    d,_ = tree.query(xyz_sta0, k=nearNum)
    sta_ensemble = sta_to[meteva.base.get_coord_names()]
    for i in range(nearNum):
        data_name = "data" + str(i)
        if nearNum ==1:
            sta_ensemble[data_name] = d[:]
        else:
            sta_ensemble[data_name] = d[:,i]
    if drop_frist:
        sta_ensemble = sta_ensemble.drop(columns=['data0'])
    return sta_ensemble

def sta_index_ensemble_near_by_grid(sta, grid,nearNum = 1):
    ER = meteva.base.ER
    members = np.arange(nearNum).tolist()
    grid1 = meteva.base.grid(grid.glon,grid.glat,member_list=members)
    grd_en = meteva.base.grid_data(grid1)
    xyz_sta =  meteva.base.tool.math_tools.lon_lat_to_cartesian(sta.loc[:,"lon"], sta.loc[:,"lat"],R = ER)
    lon = np.arange(grid1.nlon) * grid1.dlon + grid1.slon
    lat = np.arange(grid1.nlat) * grid1.dlat + grid1.slat
    grid_lon,grid_lat = np.meshgrid(lon,lat)
    xyz_grid = meteva.base.tool.math_tools.lon_lat_to_cartesian(grid_lon.flatten(), grid_lat.flatten(),R = ER)
    tree = cKDTree(xyz_sta)
    value, inds = tree.query(xyz_grid, k=nearNum)
    grd_en.values = inds.reshape((nearNum,1,1,1,grid1.nlat,grid1.nlon))
    return grd_en

def sta_values_ensemble_near_by_grid(sta, grid,nearNum = 1):
    ER = meteva.base.ER
    members = np.arange(nearNum).tolist()
    grid1 = meteva.base.grid(grid.glon,grid.glat,member_list=members)
    grd_en = meteva.base.grid_data(grid1)
    xyz_sta =  meteva.base.tool.math_tools.lon_lat_to_cartesian(sta.loc[:,"lon"], sta.loc[:,"lat"],R = ER)
    lon = np.arange(grid1.nlon) * grid1.dlon + grid1.slon
    lat = np.arange(grid1.nlat) * grid1.dlat + grid1.slat
    grid_lon,grid_lat = np.meshgrid(lon,lat)
    xyz_grid = meteva.base.tool.math_tools.lon_lat_to_cartesian(grid_lon.flatten(), grid_lat.flatten(),R = ER)
    tree = cKDTree(xyz_sta)
    value, inds = tree.query(xyz_grid, k=nearNum)
    data_name = meteva.base.get_stadata_names(sta)[0]
    input_dat = sta[data_name].values

    if nearNum ==1:
        values = input_dat[inds].reshape((grid1.nlat,grid1.nlon))
        grd_en.values[0, 0, 0, 0, :, :] = values[:,:]
    else:
        for i in range(nearNum):
            values = input_dat[inds[:,i]].reshape((grid1.nlat, grid1.nlon))
            grd_en.values[0, 0, 0, 0, :, :] = values[:, :]
            grd_en.values[i,0,0,0,:,:] = input_dat[inds[:,i]].reshape((grid1.nlat,grid1.nlon))

    return grd_en

def sta_dis_ensemble_near_by_grid(sta, grid,nearNum = 1):
    ER = meteva.base.ER
    members = np.arange(nearNum).tolist()
    grid1 = meteva.base.grid(grid.glon,grid.glat,member_list=members)
    grd_en = meteva.base.grid_data(grid1)
    xyz_sta =  meteva.base.tool.math_tools.lon_lat_to_cartesian(sta.loc[:,"lon"], sta.loc[:,"lat"],R = ER)
    lon = np.arange(grid1.nlon) * grid1.dlon + grid1.slon
    lat = np.arange(grid1.nlat) * grid1.dlat + grid1.slat
    grid_lon,grid_lat = np.meshgrid(lon,lat)
    xyz_grid = meteva.base.tool.math_tools.lon_lat_to_cartesian(grid_lon.flatten(), grid_lat.flatten(),R = ER)
    tree = cKDTree(xyz_sta)
    dis, inds = tree.query(xyz_grid, k=nearNum)
    grd_en.values = dis.reshape((nearNum,1,1,1,grid1.nlat,grid1.nlon))
    return grd_en


def ids_list_list_in_r_of_sta(sta_to, r = 40, sta_from = None,drop_first = False):
    '''

    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    sta_from = meteva.base.not_IV(sta_from)
    #print(sta_from)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    input_dat = sta_from["id"].values
    indexs_list = tree.query_ball_point(xyz_sta0,r=r)
    ids_list = []
    if drop_first:
        for indexs in indexs_list:
            if len(indexs) >1:
                values = input_dat[indexs[1:]]
                ids_list.append(values)
            else:
                ids_list.append([meteva.base.IV])
    else:
        for indexs in indexs_list:
            values = input_dat[indexs]
            ids_list.append(values)
    return ids_list

def get_stations_near_by_cyclone_trace(sta_cyclone_trace,station, r = 1000):
    '''

    :param sta_cyclone_trace: 台风轨迹，每一行记录了台风一个时刻的经纬度
    :param station:  站点表
    :param r:  查询半径
    :return:  站点数据，所有时刻相邻站点的集合
    '''
    trace_point_list, time_list = meteva.base.fun.grouping.group(sta_cyclone_trace, g="time") #将台风轨迹拆分成多个站点数据
    sta_near_list = []
    for i in range(len(trace_point_list)):
        sta_to = trace_point_list[i]  #每个时刻的台风位置是包含一行记录的站点数据
        xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:], R=meteva.base.basicdata.ER)
        xyz_sta1 = lon_lat_to_cartesian(station['lon'].values[:], station['lat'].values[:],
                                        R=meteva.base.basicdata.ER)
        tree = cKDTree(xyz_sta1)
        indexs_list = tree.query_ball_point(xyz_sta0, r=r)
        near_sta =station.iloc[indexs_list[0],:]
        near_sta["time"] = time_list[i]
        sta_near_list.append(near_sta)
    near_station_all = meteva.base.concat(sta_near_list)
    return near_station_all

def values_list_list_in_r_of_sta(sta_to, r = 40, sta_from = None,drop_first = False):
    '''

    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''
    if(sta_to is None):
        return None
    if(sta_from is None):
        sta_from = copy.deepcopy(sta_to)
    sta_from = meteva.base.not_IV(sta_from)
    #print(sta_from)
    xyz_sta0 = lon_lat_to_cartesian(sta_to['lon'].values[:], sta_to['lat'].values[:],R = meteva.base.basicdata.ER)
    xyz_sta1 = lon_lat_to_cartesian(sta_from['lon'].values[:], sta_from['lat'].values[:],R = meteva.base.basicdata.ER)
    tree = cKDTree(xyz_sta1)
    data_name = meteva.base.get_stadata_names(sta_from)[0]
    input_dat = sta_from[data_name].values
    indexs_list = tree.query_ball_point(xyz_sta0,r=r)

    values_list = []
    if drop_first:
        for indexs in indexs_list:
            if len(indexs) >1:
                values = input_dat[indexs[1:]]
                values_list.append(values)
            else:
                values_list.append([meteva.base.IV])
    else:
        for indexs in indexs_list:
            values = input_dat[indexs]
            values_list.append(values)

    return values_list

def statistic_in_r_of_sta(sta_to,operation, r = 40, sta_from = None,drop_first = False):
    '''
    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''
    sta_result = sta_to[meteva.base.get_coord_names()]
    values_list_list = values_list_list_in_r_of_sta(sta_to, r = r, sta_from = sta_from,drop_first = drop_first)
    statistic_value = []
    #print(values_list_list)
    nsta = len(values_list_list)
    for n in  range(nsta):
        values_list = values_list_list[n]
        if(len(values_list)==0):
            statistic_value.append(meteva.base.IV)
        else:
            statistic_value.append(operation(values_list))
    statistic_value = np.array(statistic_value)
    sta_result[operation.__name__] = statistic_value
    return sta_result

def max_in_r_of_sta(sta_to, r = 40, sta_from = None,drop_first = False):
    '''
    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''

    return statistic_in_r_of_sta(sta_to,np.max, r = r, sta_from = sta_from,drop_first = drop_first)

def min_in_r_of_sta(sta_to, r = 40, sta_from = None,drop_first = False):
    '''
    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''

    return statistic_in_r_of_sta(sta_to,np.min, r = r, sta_from = sta_from,drop_first = drop_first)

def mean_in_r_of_sta(sta_to, r = 40, sta_from = None,drop_first = False):
    '''
    :param sta_to:
    :param r:
    :param sta_from:
    :param drop_first:
    :return: 返回的站点将会和sta_to 一致
    '''

    return statistic_in_r_of_sta(sta_to,np.mean, r = r, sta_from = sta_from,drop_first = drop_first)


def add_stavalue_to_nearest_grid(sta,grid):
    '''

    :param sta:
    :param grid:
    :return:
    '''

    grd = meteva.base.grid_data(meteva.base.grid(grid.glon,grid.glat,gtime=[sta.iloc[0,1]],dtime_list=[sta.iloc[0,2]],level_list=[sta.iloc[0,0]]))
    sta1 = meteva.base.sele.in_grid_xy(sta, grid)
    data_names = meteva.base.get_stadata_names(sta1)
    ig = np.round((sta1.loc[:,'lon'].values - grid.slon) // grid.dlon).astype(dtype = 'int16')
    jg = np.round((sta1.loc[:,'lat'].values - grid.slat) // grid.dlat).astype(dtype = 'int16')
    df = pd.DataFrame({"jg":jg,"ig":ig,"value":sta1.iloc[:,-1]})
    duplicate_data_sum = df.groupby(by=['jg','ig'],as_index=False)["value"].sum()
    ig_d = duplicate_data_sum["ig"].values
    jg_d = duplicate_data_sum["jg"].values
    value = duplicate_data_sum["value"].values
    grd.values[0, 0, 0, 0, jg_d, ig_d] = value[:]
    meteva.base.set_griddata_coords(grd, member_list=data_names)
    return grd

def add_stacount_to_nearest_grid(sta,grid):
    sta1 = sta.copy()
    data_names = meteva.base.get_stadata_names(sta)
    sta1 = meteva.base.sele_by_para(sta1,member=data_names[0])
    sta1.iloc[:,-1] = 1
    return add_stavalue_to_nearest_grid(sta1,grid)





# -*- coding: utf-8 -*-
"""
Created on Fri Jul 10 14:58:07 2020

@author: 1
"""

#============================  FeatureFinder()  ==============================
#Function:Identify spatial features within a verification set using a threshold-based method.

#from skimage import measure,color
import numpy as np
import sys
import meteva
import math
from scipy.ndimage import convolve
from queue import Queue

def measure_label(arr0):
    arr = np.zeros((arr0.shape[0]+2,arr0.shape[1]+2)) > 1
    arr[1:-1,1:-1] = arr0[:,:]
    nx = arr.shape[1]
    ny = arr.shape[0]
    label = np.zeros(arr.shape)
    num = 0
    for j in range(1, ny - 1):
        for i in range(1,nx-1):
            if arr[j,i] and label[j,i] == 0:
                num = num + 1
                label[j,i] = num
                qu = Queue(maxsize=0)
                qu.put([j,i])
                while not qu.empty():
                    ji = qu.get()
                    for p in range(-1,2):
                        for q in range(-1,2):
                            if label[ji[0]+p,ji[1]+q] ==0 and arr[ji[0]+p,ji[1]+q]:
                                qu.put([ji[0]+p,ji[1]+q])
                                label[ji[0] + p, ji[1] + q] = num

    label_c = label[1:-1,1:-1]

    return label_c


def get_disk_kernel(r):
    r2 = r * r
    rint = math.ceil(r - 0.5)
    diskg = np.arange(-rint, rint + 1, 1)
    n = len(diskg)
    x, y = np.meshgrid(diskg, diskg)
    xy = np.array([x, y])
    abmax = np.max(np.abs(xy), axis=0)
    abmin = np.min(np.abs(xy), axis=0)
    hold = np.full((n, n), np.nan)
    tmp1 = np.full((n, n), np.nan)
    tmp2 = np.full((n, n), np.nan)
    val = (abmax + 0.5) ** 2 + (abmin - 0.5) ** 2
    id1 = (r2 < val)
    id2 = (r2 >= val)

    tmp1[id1] = abmin[id1] - 0.5
    tmp1[id2] = np.sqrt(r2 - (abmax[id2] + 0.5) ** 2)
    val = (abmax - 0.5) ** 2 + (abmin + 0.5) ** 2

    id1 = r2 < val
    id2 = r2 >= val

    tmp2[id1] = abmin[id1] + 0.5
    tmp2[id2] = np.sqrt(r2 - (abmax[id2] - 0.5) ** 2)

    val1 = (abmax + 0.5) ** 2 + (abmin + 0.5) ** 2
    hold_id1 = (r2 < val1) & (r2 > ((abmax - 0.5) ** 2 + (abmin - 0.5) ** 2))
    hold_id2 = ((abmin == 0) & (abmax - 0.5 < r) & (abmax + 0.5 >= r))
    hold_id = hold_id1 | hold_id2

    hold[hold_id] = (r2 * (0.5 * (np.arcsin(tmp2[hold_id] / r) - np.arcsin(tmp1[hold_id] / r))
                           + 0.25 * (np.sin(2 * np.arcsin(tmp2[hold_id] / r)) - np.sin(
                2 * np.arcsin(tmp1[hold_id] / r)))
                           )
                     - (abmax[hold_id] - 0.5) * (tmp2[hold_id] - tmp1[hold_id])
                     + (tmp1[hold_id] - abmin[hold_id] + 0.5))

    hold[np.isnan(hold)] = 0
    hold = hold + (val1 < r2)
    hold[rint, rint] = min(math.pi * r2, math.pi / 2)
    rc2 = rint - 0.5
    if (rint > 0) and (r > rc2) and (r2 < (rc2 ** 2 + 0.25)):
        tmp1 = math.sqrt(r2 - rc2 ** 2)
        tmp1n = tmp1 / r
        hold0 = 2 * (r2 * (0.5 * np.arcsin(tmp1n) + 0.25 * np.sin(2 * np.arcsin(tmp1n))) - tmp1 * (rint - 0.5))
        hold[2 * rint, rint] = hold0
        hold[rint, 2 * rint] = hold0
        hold[rint, 0] = hold0
        hold[0, rint] = hold0
        hold[2 * rint - 1, rint] = hold[2 * rint - 1, rint] - hold0
        hold[rint, 2 * rint - 1] = hold[rint, 2 * rint - 1] - hold0
        hold[rint, 1] = hold[rint, 1] - hold0
        hold[1, rint] = hold[1, rint] - hold0
    hold[rint, rint] = min(hold[rint, rint], 1)
    sn = np.sum(hold)
    kernel = hold/sn
    return kernel

def get_disk_kernel_1(r):
    x = np.arange(-r,r+1)
    y = np.arange(-r,r+1)
    xx,yy = np.meshgrid(x,y)
    DD = r**2 - xx**2 - yy**2+1
    DD[DD<0] = 0
    DD[DD>0] = 1
    sum1 = np.sum(DD)
    kernel = DD/sum1
    return kernel


#start = time.clock()    #开始时间

#连通域标记
# def labelsConnection(data):
#     labels=measure.label(data,connectivity=2)  #connectivity表示连接的模式，1代表4连通，2代表8连通
#     meteva.base.mesh(labels,save_path=r"H:\test_data\output\method\space\mode\label.png")
#     dst=color.label2rgb(labels)  #根据不同的标记显示不同的颜色
#     #print('regions number:',labels.max()+1)  #显示连通区域块数(从0开始标记)
#
#     return dst

#每个连通域内格点数统计,并且判断格点数量是否在阈值范围内
def propCounts(data_2, minsize, maxsize):
    labels = measure_label(data_2)
    #labels = measure.label(data_2,connectivity=2)

    valid_label = []
    area = []
    current_bw = np.array(())
    max_label = int(np.max(labels))
    labelsfeature = {}
    # for prop in properties:
    #     nums.append(prop.area)    #统计连通域内的格点数
    #
    # numsID = list(map(lambda x :any(x > [minsize]) and any(x < [maxsize]), nums))    #比较格点数是否在阈值范围内

    nums = []
    numsID = []
    for i in range(1,max_label+1):
        index = np.where(labels ==i)
        count = len(index[0])
        nums.append(count)
        numsID.append(count>minsize and count<maxsize)
    for i in range(len(numsID)):
        if numsID[i]:
            valid_label_0 = i
            area_0 = nums[i]
            area.append(area_0)
            #print(numsID[i], valid_label_0)
            valid_label.append(valid_label_0)
            current_bw = np.in1d(labels, list(valid_label)).reshape(labels.shape)
            #labels = prop._label_image    #读取的prop._label_image数据是已经进行过旋转的结果


    for i in range(len(valid_label)):
        n = valid_label[i]
        #print(n)
        #f = np.where(labels, labels == n + 1, 0)
        #print(f.size)
        f = np.where(labels == n+1)
        f = {(i + 1):f}
        labelsfeature.update(f)
        #添加属性
        xrange = [0.5, np.shape(data_2)[1] + 0.5]
        yrange = [0.5, np.shape(data_2)[0] + 0.5]
        dim = np.shape(data_2)
        xcol = np.arange(1, np.shape(data_2)[1] + 1, 1)
        yrow = np.arange(1, np.shape(data_2)[0] + 1, 1)

        warnings = ["Row index corresponds to increasing y coordinate; column to increasing x",
                    "Transpose matrices to get the standard presentation in R",
                    "Example: image(result$xcol,result$yrow,t(result$d))"]
        labelsfeature.update({'Type':'mask', 'xrange': xrange, 'yrange': yrange,
                              'dim': dim, 'xstep': 1, 'ystep': 1, 'area': area, 
                              'warnings': warnings, 'xcol':xcol, 'yrow':yrow})
    labelsfeature["label_count"]=len(valid_label)
    #print(labelsfeature)
    properties = None
    return nums, numsID, current_bw, properties, labelsfeature

#R里面新定义的Nfun函数，求和函数,实际函数功能是统计连通域内面积大小,python用prop.area代替
def newFun(Obj):
    #axis为:1：计算每一行的和，axis为0：计算每一列的和,keepdims为保持原有维度
    return np.sum(np.sum(Obj, axis=0, keepdims=True))

#
# def feature_finder_bak(grd_ob, grd_fo, smooth, threshold, minsize, maxsize=float("Inf"),
#                    do_smooth=True):
#     # 读观测数据
#     # dataset_ob = xr.open_dataset(filename_ob)  #通过xarray程序库读取nc文件中的所有内容
#     # X = np.squeeze(grd_ob.variables['data0'])
#
#     # 读预报数据
#     # dataset_fcst = xr.open_dataset(filename_fo)  #通过xarray程序库读取nc文件中的所有内容
#     # Xhat = np.squeeze(grd_fo.variables['data0'])
#
#     X = np.squeeze(np.array(grd_ob))
#     Xhat = np.squeeze(np.array(grd_fo))
#
#     # 读经纬度，并形成R里面的格式
#     lon = grd_ob['lon']
#     lat = grd_ob['lat']
#     X_lon, Y_lat = np.meshgrid(lon, lat)
#     loc1 = X_lon.reshape(X_lon.size, order='F')
#     loc2 = Y_lat.reshape(Y_lat.size, order='F')
#     loc = np.vstack((loc1, loc2)).T
#
#     Object = {"grd_ob": X, "grd_fo": Xhat, "loc": loc}  # hold是make_saptialVx的计算结果
#     thresh = threshold
#     smoothpar = smooth
#     smoothfun = "disk2dsmooth"
#     smoothfunargs = None
#     idfun = "disjointer"
#     zerodown = False
#     timepoint = 1
#     obs = 1
#     model = 1
#     fac = 1
#
#     if type(minsize) == list:
#         minsize = np.array(minsize)
#     if type(maxsize) == list:
#         maxsize = np.array(maxsize)
#     if type(smooth) == list:
#         smooth = np.array(smooth)
#     if type(thresh) == list:
#         thresh = np.array(thresh)
#
#         # theCall <- match.call()    #调用match函数
#
#     if (np.size(minsize) == 1):
#         minsize = np.tile(minsize, 2)  # 输入观测和预报数据，因而将最小值和最大值展开为两个数
#     if (np.size(maxsize) == 1):  # error:object of type 'float' has no len()
#         maxsize = np.tile(maxsize, 2)
#
#     if (len(minsize) != 2):
#         try:
#             sys.exit(0)
#         except:
#             print("FeatureFinder: invalid min.size argument.  Must have length one or two.")
#     if (len(maxsize) != 2):
#         try:
#             sys.exit(0)
#         except:
#             print("FeatureFinder: invalid max.size argument.  Must have length one or two.")
#     if (any(minsize) < 1):  # a为list
#         try:
#             sys.exit(0)
#         except:
#             print("FeatureFinder: invalid min.size argument.  Must be >= 1.")
#     if (any(maxsize) < any(minsize)):
#         try:
#             sys.exit(0)
#         except:
#             print("FeatureFinder: invalid max.size argument.  Must be >= min.size argument.")
#     a = Object.copy()
#     dat = {'ob': Object['grd_ob'], 'fo': Object['grd_fo']}  # 引用hold里的X,Xhat要素并赋值
#     X = dat['ob']
#     Y = dat['fo']
#     xdim = X.shape
#
#     if (do_smooth):
#         if (np.size(smoothpar) == 1):
#             smoothpar = np.tile(smoothpar, 2)  # 观测场和预报场的平滑参数一致
#         elif (len(smoothpar) > 2):
#             try:
#                 sys.exit(0)
#             except:
#                 print("FeatureFinder: invalid smoothpar argument.  Must have length one or two.")
#
#         # 调用disk2dsmooth中的kernel2dsmooth卷积平滑，python里面目前用2D卷积平滑替代
#         kernel_X = np.ones((smoothpar[0], smoothpar[0]), np.float32) / 5  # X的卷积核
#         kernel_Y = np.ones((smoothpar[1], smoothpar[1]), np.float32) / 5  # Y的卷积核
#         #Xsm = cv.filter2D(np.rot90(X, 4), -1, kernel_X)  # 对X做2D卷积平滑,旋转4次=没有旋转，不做旋转会报错（opencv版本问题）
#         Xsm = convolve(X, kernel_X)
#         #Ysm = cv.filter2D(np.rot90(Y, 4), -1, kernel_Y)  # 对Y做2D卷积平滑,旋转4次=没有旋转，不做旋转会报错（opencv版本问题）
#         Ysm = convolve(Y, kernel_Y)
#         if (zerodown):
#             Xsm = np.where(Xsm > 0, Xsm, 0)  # Xsm中大于0的值被0代替
#             Ysm = np.where(Ysm > 0, Ysm, 0)  # Ysm中大于0的值被0代替
#     else:
#         Xsm = X
#         Ysm = Y
#
#     if (np.size(thresh) == 1):
#         thresh = np.tile(thresh, 2)
#     thresh = thresh * fac
#     # 二值化，首先生成0矩阵，然后将大于阈值的部分赋值为1
#     # 但是python里面在进行连通域分析的时候已经进行过二值化，不必单独进行二值化
#     # sIx = np.zeros((xdim[0], xdim[1]))
#     # sIy = np.zeros((xdim[0], xdim[1]))
#     # sIx = np.where(Xsm < thresh[0], Xsm, 1)
#     # sIy = np.where(Ysm < thresh[1], Xsm, 1)
#
#     # 连通域分析
#     # 连通域分析的时候，分析的是经过模糊处理的图像
#     # R里面的阈值为5,2D卷积平滑时增强图像，阈值设为310才能结果对应
#     Xfeats = labelsConnection(Xsm > thresh[0])
#     Yfeats = labelsConnection(Ysm > thresh[1])
#     if (len(Xfeats) == 0):
#         Xfeats = None
#     if (len(Yfeats) == 0):
#         Yfeats = None
#     # 如果对连通域的面积大小做限制，则需要执行下面的判断
#     if (any(minsize > 1) or any(maxsize < X.size)):
#         # 统计每个连通域内的格点数
#         if (np.all(Xfeats) != None):
#             Xnums = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[0]
#             XnumsID = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[1]
#             Xfeats = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[2]
#             Xprop = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[3]
#
#         if (np.all(Yfeats) != None):
#             Ynums = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[0]
#             YnumsID = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[1]
#             Yfeats = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[2]
#             Yprop = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[3]
#
#     # Xlab = np.zeros(xdim[0], xdim[1])
#     # Ylab = np.zeros(xdim[0], xdim[1])
#
#     if (np.all(Xfeats) != None):
#         Xlab = Xfeats
#     else:
#         Xfeats = None
#
#     if (np.all(Yfeats) != None):
#         Ylab = Yfeats
#     else:
#         Yfeats = None
#
#         # 返回的out列表里面包括:观测数据（X）,预报数据（Xhat），连通域分析后单个区域（X.feats,Yfeats）并包括了很多的属性信息
#     # X.feats,Yfeats包括了很多的属性信息：x,y的范围xrange,yrange，维度dim,步长xstep,ystep,行列信息等；标记后的连通域（Xlab,Ylab）；
#     # 识别函数、标记函数名称：Convolution Threshold。
#     Xprop = propCounts(Xsm > thresh[0], minsize[0], maxsize[0])[3]
#     Yprop = propCounts(Ysm > thresh[1], minsize[1], maxsize[1])[3]
#     Xlabelsfeature = propCounts(Xsm > thresh[0], minsize[0], maxsize[0])[4]
#     Ylabelsfeature = propCounts(Ysm > thresh[1], minsize[1], maxsize[1])[4]
#
#     '''
#     out = {'ob':Object['ob'], 'fo':Object['fo'], 'loc':Object['loc'], "Xlabeled":Xfeats, "Ylabeled":Yfeats,
#            "identifier_function" : "convthresh", "identifier_label" : "Convolution Threshold",
#            "attr_timepoint" : "time_point", "attr_model": "model", "attr_call": "theCall",
#            "Xprop":Xprop, "Yprop":Yprop, "Xlabelsfeature":Xlabelsfeature, "Ylabelsfeature":Ylabelsfeature}
#     '''
#     out = {'grd_ob': Object['grd_ob'], 'grd_fo': Object['grd_fo'], 'loc': Object['loc'],
#            "grd_ob_labeled": Xfeats, "grd_fo_labeled": Yfeats,
#            "identifier_label": "Convolution Threshold",
#            "grd_ob_prop": Xprop, "grd_fo_prop": Yprop,
#            "grd_ob_features": Xlabelsfeature, "grd_fo_features": Ylabelsfeature}
#
#     Xlabels = data_pre.pick_labels(copy.deepcopy(Xlabelsfeature))
#     xtmp = Xlabels.copy()
#     Xlabeled = data_pre.relabeled(xtmp)
#     Ylabels = data_pre.pick_labels(copy.deepcopy(Ylabelsfeature))
#     ytmp = Ylabels.copy()
#     Ylabeled = data_pre.relabeled(ytmp)
#     out.update({"grd_ob_labeled": Xlabeled, "grd_fo_labeled": Ylabeled})
#     # 每个格点的长、宽
#     a = np.unique(out['loc'][:, 0])[1] - np.unique(out['loc'][:, 0])[0]  # 第一列为经度间隔
#     b = np.unique(out['loc'][:, 1])[1] - np.unique(out['loc'][:, 1])[0]  # 第二列为维度间隔
#     S = a * b
#     ob_area = (np.array(out['grd_ob_features']['area']) * S).tolist()
#     out['grd_ob_features']['area'] = ob_area
#
#     print(out['grd_fo_features'])
#     fo_area = (np.array(out['grd_fo_features']['area']) * S).tolist()
#     out['grd_fo_features']['area'] = fo_area
#     out["grid"] = meteva.base.get_grid_of_data(grd_ob)
#     return out

def feature_finder(grd_ob0, grd_fo0, smooth, threshold, minsize,compare = ">=",maxsize = float("Inf")):
    
    #读观测数据
    #dataset_ob = xr.open_dataset(filename_ob)  #通过xarray程序库读取nc文件中的所有内容
    #X = np.squeeze(grd_ob.variables['data0'])
    
    #读预报数据
    #dataset_fcst = xr.open_dataset(filename_fo)  #通过xarray程序库读取nc文件中的所有内容
    #Xhat = np.squeeze(grd_fo.variables['data0'])

    grd_ob = grd_ob0.copy()
    grd_ob.attrs["var_name"] = "原始场"
    grd_fo = grd_fo0.copy()
    grd_fo.attrs["var_name"] = "原始场"

    X = np.squeeze(np.array(grd_ob)).astype(np.float32)
    Xhat = np.squeeze(np.array(grd_fo)).astype(np.float32)
    
    #读经纬度，并形成R里面的格式
    lon = grd_ob['lon']
    lat = grd_ob['lat']    
    X_lon, Y_lat = np.meshgrid(lon, lat)
    loc1 = X_lon.reshape(X_lon.size, order = 'F')
    loc2 = Y_lat.reshape(Y_lat.size, order = 'F')
    loc = np.vstack((loc1, loc2)).T    
    
    Object = {"grd_ob": X, "grd_fo": Xhat, "loc":loc}    #hold是make_saptialVx的计算结果
    thresh = threshold


    smoothfun = "disk2dsmooth"
    smoothfunargs = None
    idfun = "disjointer"
    zerodown = False
    timepoint = 1
    obs = 1
    model = 1
    fac = 1

    if not isinstance(smooth,list):
        smooth = [smooth,smooth]

    if not isinstance(minsize,list):
        minsize = [minsize,minsize]

    if not isinstance(maxsize,list):
        maxsize = [maxsize,maxsize]

    if not isinstance(thresh,list):
        thresh = [thresh,thresh]


    
    #theCall <- match.call()    #调用match函数

    if (np.size(minsize) == 1):
        minsize = np.tile(minsize, 2)        #输入观测和预报数据，因而将最小值和最大值展开为两个数
    if (np.size(maxsize) == 1):    #error:object of type 'float' has no len()
        maxsize = np.tile(maxsize, 2)
        
    if (len(minsize) != 2):
        try:
            sys.exit(0)
        except:
            print("FeatureFinder: invalid min.size argument.  Must have length one or two.")
    if (len(maxsize) != 2):
        try:
            sys.exit(0)
        except:
            print("FeatureFinder: invalid max.size argument.  Must have length one or two.")
    if (any(minsize) < 1):    #a为list
        try:
            sys.exit(0)
        except:
            print("FeatureFinder: invalid min.size argument.  Must be >= 1.")
    if (any(maxsize) < any(minsize)):
        try:
            sys.exit(0)
        except:
            print("FeatureFinder: invalid max.size argument.  Must be >= min.size argument.")
    a = Object.copy() 
    dat = {'ob':Object['grd_ob'], 'fo':Object['grd_fo']}    #引用hold里的X,Xhat要素并赋值
    X = dat['ob']
    Y = dat['fo']
    xdim = X.shape

    if not isinstance(smooth,list):
        smooth = [smooth,smooth]

    if smooth[0]>0:

        kernel_X = get_disk_kernel(smooth[0])
        #Xsm = cv.filter2D(np.rot90(X, 4), -1, kernel_X)    #对X做2D卷积平滑,旋转4次=没有旋转，不做旋转会报错（opencv版本问题）
        #Xsm = cv.filter2D(X, -1, kernel_X)  # 对X做2D卷积平滑,旋转4次=没有旋转，不做旋转会报错（opencv版本问题）

        Xsm = convolve(X, kernel_X)
        if (zerodown):
             Xsm = np.where(Xsm > 0, Xsm, 0)    #Xsm中大于0的值被0代替
    else:
        Xsm = X

    grd_ob_smooth = grd_ob.copy()
    grd_ob_smooth.attrs["var_name"] = "平滑场"
    grd_ob_smooth[:] = Xsm[:]


    if smooth[1] >0:
        kernel_Y = get_disk_kernel(smooth[1])
        #Ysm = cv.filter2D(np.rot90(Y, 4), -1, kernel_Y)  # 对Y做2D卷积平滑,旋转4次=没有旋转，不做旋转会报错（opencv版本问题）
        Ysm = convolve(Y, kernel_Y)
        if (zerodown):
            Ysm = np.where(Ysm > 0, Ysm, 0)  # Ysm中大于0的值被0代替
    else:
        Ysm = Y

    grd_fo_smooth = grd_fo.copy()
    grd_fo_smooth.attrs["var_name"] = "平滑场"
    grd_fo_smooth[:] = Ysm[:]

    if (np.size(thresh) == 1):
        thresh = np.tile(thresh, 2)
    thresh = thresh * fac
    #二值化，首先生成0矩阵，然后将大于阈值的部分赋值为1
    #但是python里面在进行连通域分析的时候已经进行过二值化，不必单独进行二值化
    #sIx = np.zeros((xdim[0], xdim[1]))
    #sIy = np.zeros((xdim[0], xdim[1]))
    #sIx = np.where(Xsm < thresh[0], Xsm, 1)
    #sIy = np.where(Ysm < thresh[1], Xsm, 1)
    
    #连通域分析
    #连通域分析的时候，分析的是经过模糊处理的图像
    #R里面的阈值为5,2D卷积平滑时增强图像，阈值设为310才能结果对应
    # Xfeats = labelsConnection(Xsm > thresh[0])
    # Yfeats = labelsConnection(Ysm > thresh[1])
    # if (len(Xfeats) == 0):
    #     Xfeats = None
    # if (len(Yfeats) == 0):
    #     Yfeats = None
    #如果对连通域的面积大小做限制，则需要执行下面的判断


    # if minsize[0] > 1 or maxsize[0] < X.size:
    #     #统计每个连通域内的格点数
    #     if (np.all(Xfeats) != None):
    #         Xnums = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[0]
    #         XnumsID = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[1]
    #         Xfeats = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[2]
    #         Xprop = propCounts(np.rot90(Xsm > thresh[0], 4), minsize[0], maxsize[0])[3]
    #
    # if minsize[1] > 1 or minsize[1] < X.size:
    #     if (np.all(Yfeats) != None):
    #         Ynums = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[0]
    #         YnumsID = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[1]
    #         Yfeats = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[2]
    #         Yprop = propCounts(np.rot90(Ysm > thresh[1], 4), minsize[1], maxsize[1])[3]

    
    #Xlab = np.zeros(xdim[0], xdim[1])
    #Ylab = np.zeros(xdim[0], xdim[1])
    
    # if (np.all(Xfeats) != None):
    #     Xlab = Xfeats
    # else:
    #     Xfeats = None
    #
    # if (np.all(Yfeats) != None):
    #     Ylab = Yfeats
    # else:
    #     Yfeats = None
    
    #返回的out列表里面包括:观测数据（X）,预报数据（Xhat），连通域分析后单个区域（X.feats,Yfeats）并包括了很多的属性信息
    #X.feats,Yfeats包括了很多的属性信息：x,y的范围xrange,yrange，维度dim,步长xstep,ystep,行列信息等；标记后的连通域（Xlab,Ylab）；

    ob_hp = None
    fo_hp = None
    if compare[0] ==">=":
        ob_hp = Xsm >= thresh[0]
        fo_hp = Ysm >= thresh[1]
    elif compare[0] ==">":
        ob_hp = Xsm > thresh[0]
        fo_hp = Ysm > thresh[1]
    elif compare[0] =="<=":
        ob_hp = Xsm <= thresh[0]
        fo_hp = Ysm <= thresh[1]
    elif compare[0] =="<":
        ob_hp = Xsm <= thresh[0]
        fo_hp = Ysm <= thresh[1]
    else:
        print("compare parameter must be >=  >  <  or <=" )



    # #识别函数、标记函数名称：Convolution Threshold。
    # Xprop = propCounts(Xsm > thresh[0], minsize[0], maxsize[0])[3]
    # Yprop = propCounts(Ysm > thresh[1], minsize[1], maxsize[1])[3]
    # Xlabelsfeature = propCounts(Xsm > thresh[0], minsize[0], maxsize[0])[4]
    # Ylabelsfeature = propCounts(Ysm > thresh[1], minsize[1], maxsize[1])[4]

    Xprop = propCounts(ob_hp, minsize[0], maxsize[0])[3]
    Yprop = propCounts(fo_hp, minsize[1], maxsize[1])[3]
    Xlabelsfeature = propCounts(ob_hp, minsize[0], maxsize[0])[4]
    Ylabelsfeature = propCounts(fo_hp, minsize[1], maxsize[1])[4]

    out = {'grd_ob':grd_ob, 'grd_fo':grd_fo,
           "grd_ob_smooth":grd_ob_smooth,"grd_fo_smooth":grd_fo_smooth,
           "grd_ob_label":None, "grd_fo_label":None,
           "identifier_label" : "Convolution Threshold",
           "grd_ob_prop":Xprop, "grd_fo_prop":Yprop,
           "grd_ob_features":Xlabelsfeature, "grd_fo_features":Ylabelsfeature}

    #print(Xlabelsfeature)
    #Xlabels = data_pre.pick_labels(copy.deepcopy(Xlabelsfeature))
    grid0 = meteva.base.get_grid_of_data(grd_ob)


    grd_ob_labeled = grd_ob.copy()
    grd_ob_labeled.attrs["var_name"] = "目标编号"
    label_value = np.zeros((grid0.nlat,grid0.nlon))
    for i in range(Xlabelsfeature["label_count"]):
        label = Xlabelsfeature[i + 1]
        label_value[label] = i+1
    grd_ob_labeled.values[:] = label_value[:]

    grd_fo_labeled = grd_fo.copy()
    grd_fo_labeled.attrs["var_name"] = "目标编号"
    label_value = np.zeros((grid0.nlat,grid0.nlon))
    for i in range(Ylabelsfeature["label_count"]):
        label = Ylabelsfeature[i + 1]
        label_value[label] = i+1
    grd_fo_labeled.values[:] = label_value[:]
    #grd_fo_labeled.values[:] = Ylabeled[:]

    out.update({ "grd_ob_label":grd_ob_labeled, "grd_fo_label":grd_fo_labeled})
    #每个格点的长、宽

    #a = np.unique(out['loc'][:, 0])[1]-np.unique(out['loc'][:, 0])[0]    #第一列为经度间隔
    #b = np.unique(out['loc'][:, 1])[1]-np.unique(out['loc'][:, 1])[0]    #第二列为维度间隔
    a = grid0.dlon
    b = grid0.dlat
    S = a * b
    #if len(out["grd_ob_features"].keys())>0:
    if out["grd_ob_features"]["label_count"] >0:
            ob_area = (np.array(out['grd_ob_features']['area'])*S).tolist()
            out['grd_ob_features']['area'] = ob_area
    #if len(out["grd_fo_features"].keys())>0:
    if out["grd_fo_features"]["label_count"] > 0:
            fo_area = (np.array(out['grd_fo_features']['area'])*S).tolist()
            out['grd_fo_features']['area'] = fo_area
    out["grid"] = grid0
    out["label_list_ob"] = np.arange(1,Xlabelsfeature["label_count"] + 1).tolist()
    out["label_list_fo"]=np.arange(1,Ylabelsfeature["label_count"] + 1).tolist()

    return out



def feature_finder_and_merge(grd_ob0,smooth, threshold, minsize,compare = ">=",near_dis = 100,near_rate = 0.3):
    grd_ob = grd_ob0.copy()
    grd_ob.attrs["var_name"] = "原始场"
    X = np.squeeze(np.array(grd_ob)).astype(np.float32)

    # 读经纬度，并形成R里面的格式
    lon = grd_ob['lon']
    lat = grd_ob['lat']
    X_lon, Y_lat = np.meshgrid(lon, lat)
    loc1 = X_lon.reshape(X_lon.size, order='F')
    loc2 = Y_lat.reshape(Y_lat.size, order='F')
    loc = np.vstack((loc1, loc2)).T

    Object = {"grd_ob": X, "loc": loc}  # hold是make_saptialVx的计算结果
    thresh = threshold
    zerodown = False
    dat = {'ob': Object['grd_ob']}  # 引用hold里的X,Xhat要素并赋值
    X = dat['ob']

    if smooth > 0:
        kernel_X = get_disk_kernel(smooth)
        Xsm = convolve(X, kernel_X)
        if (zerodown):
            Xsm = np.where(Xsm > 0, Xsm, 0)  # Xsm中大于0的值被0代替
    else:
        Xsm = X

    grd_ob_smooth = grd_ob.copy()
    grd_ob_smooth.attrs["var_name"] = "平滑场"
    grd_ob_smooth[:] = Xsm[:]

    if (np.size(thresh) == 1):
        thresh = np.tile(thresh, 2)
    fac = 1
    thresh = thresh * fac

    ob_hp = None
    if compare == ">=":
        ob_hp = Xsm >= thresh[0]
    elif compare== ">":
        ob_hp = Xsm > thresh[0]
    elif compare == "<=":
        ob_hp = Xsm <= thresh[0]
    elif compare == "<":
        ob_hp = Xsm <= thresh[0]
    else:
        print("compare parameter must be >=  >  <  or <=")

    Xprop = propCounts(ob_hp, minsize, 1e8 )[3]
    Xlabelsfeature = propCounts(ob_hp, minsize, 1e8)[4]
    grid0 = meteva.base.get_grid_of_data(grd_ob)
    if near_dis >0:
        near_dis = near_dis/111/grid0.dlon
        Xlabelsfeature = meteva.method.mode.consistent.combined_near_labels(Xlabelsfeature,near_dis,near_rate)


    out = {'grd': grd_ob,
           "grd_smooth": grd_ob_smooth,
           "grd_label": None,
           "identifier_label": "Convolution Threshold",
           "grd_prop": Xprop,
           "grd_features": Xlabelsfeature}


    grd_ob_labeled = grd_ob.copy()
    grd_ob_labeled.attrs["var_name"] = "目标编号"
    label_value = np.zeros((grid0.nlat, grid0.nlon))
    id_list = []
    for i in range(Xlabelsfeature["label_count"]):
        label = Xlabelsfeature[i + 1]
        label_value[label] = i + 1
        id_list.append(i+1)

    grd_ob_labeled.values[:] = label_value[:]

    out.update({"grd_label": grd_ob_labeled})
    # 每个格点的长、宽
    a = grid0.dlon
    b = grid0.dlat
    S = a * b
    # if len(out["grd_ob_features"].keys())>0:
    if out["grd_features"]["label_count"] > 0:
        ob_area = (np.array(out['grd_features']['area']) * S).tolist()
        out['grd_features']['area'] = ob_area
    out["grid"] = grid0
    out["id_list"] = id_list


    return out

def feature_finder_single(grd_fo0,smooth, threshold, minsize,compare = ">="):
    grd_fo = grd_fo0.copy()
    grd_fo.attrs["var_name"] = "原始场"

    X = np.squeeze(np.array(grd_fo)).astype(np.float32)

    # 读经纬度，并形成R里面的格式
    lon = grd_fo['lon']
    lat = grd_fo['lat']
    X_lon, Y_lat = np.meshgrid(lon, lat)
    loc1 = X_lon.reshape(X_lon.size, order='F')
    loc2 = Y_lat.reshape(Y_lat.size, order='F')
    loc = np.vstack((loc1, loc2)).T

    Object = {"grd": X, "loc": loc}  # hold是make_saptialVx的计算结果
    thresh = threshold
    zerodown = False

    dat = {'fo': Object['grd']}  # 引用hold里的X,Xhat要素并赋值

    Y = dat['fo']

    if smooth > 0:
        kernel_X = get_disk_kernel(smooth)
        Xsm = convolve(Y, kernel_X)
        if (zerodown):
            Xsm = np.where(Xsm > 0, Xsm, 0)  # Xsm中大于0的值被0代替
    else:
        Xsm = Y

    grd_fo_smooth = grd_fo.copy()
    grd_fo_smooth.attrs["var_name"] = "平滑场"
    grd_fo_smooth[:] = Xsm[:]

    if (np.size(thresh) == 1):
        thresh = np.tile(thresh, 2)
    fac = 1
    thresh = thresh * fac

    ob_hp = None
    if compare == ">=":
        ob_hp = Xsm >= thresh[0]
    elif compare== ">":
        ob_hp = Xsm > thresh[0]
    elif compare == "<=":
        ob_hp = Xsm <= thresh[0]
    elif compare == "<":
        ob_hp = Xsm <= thresh[0]
    else:
        print("compare parameter must be >=  >  <  or <=")

    Xprop = propCounts(ob_hp, minsize, 1e8 )[3]
    Xlabelsfeature = propCounts(ob_hp, minsize, 1e8)[4]

    out = {'grd': grd_fo,
           "grd_smooth": grd_fo_smooth,
           "grd_label": None,
           "identifier_label": "Convolution Threshold",
           "grd_prop": Xprop,
           "grd_features": Xlabelsfeature}

    # print(Xlabelsfeature)
    # Xlabels = data_pre.pick_labels(copy.deepcopy(Xlabelsfeature))
    grid0 = meteva.base.get_grid_of_data(grd_fo)

    grd_fo_labeled = grd_fo.copy()
    grd_fo_labeled.attrs["var_name"] = "目标编号"
    label_value = np.zeros((grid0.nlat, grid0.nlon))
    id_list = []
    for i in range(Xlabelsfeature["label_count"]):
        label = Xlabelsfeature[i + 1]
        label_value[label] = i + 1
        id_list.append(i+1)
    grd_fo_labeled.values[:] = label_value[:]

    # grd_fo_labeled.values[:] = Ylabeled[:]

    out.update({"grd_label": grd_fo_labeled})
    # 每个格点的长、宽

    a = grid0.dlon
    b = grid0.dlat
    S = a * b
    # if len(out["grd_fo_features"].keys())>0:
    if out["grd_features"]["label_count"] > 0:
        ob_area = (np.array(out['grd_features']['area']) * S).tolist()
        out['grd_features']['area'] = ob_area

    out["grid"] = grid0
    out["id_list"] = id_list

    return out





















# -*- coding: utf-8 -*-
########################################################################################################################
# ### This module enables you to maskout the unneccessary data outside the
#                     interest region on a matplotlib-plotted output instance
# ################### in an effecient way,You can use this script for free   ###########################################
# ######################################################################################################################
# ####USAGE:INPUT include 'originfig':the matplotlib instance##
#                         'ax': the Axes instance
#                         'shapefile': the shape file used for generating a basemap A
#                         'region':the name of a region of on the basemap A,outside the region the data is to be maskout
#           OUTPUT    is  'clip' :the the masked-out or clipped matplotlib instance.
import shapefile
from matplotlib.path import Path
from matplotlib.patches import PathPatch
import pkg_resources
import numpy as np

def getPathFromShp(shpfile, region):
    try:
        sf = shapefile.Reader(shpfile)
        vertices = []  # 这块是已经修改的地方
        codes = []  # 这块是已经修改的地方

        for shape_rec in sf.shapeRecords():
            # if shape_rec.record[3] == region:  # 这里需要找到和region匹配的唯一标识符，record[]中必有一项是对应的。
            if region == [100000] or shape_rec.record[4] in region:  # 这块是已经修改的地方
                pts = shape_rec.shape.points
                prt = list(shape_rec.shape.parts) + [len(pts)]
                for i in range(len(prt) - 1):
                    for j in range(prt[i], prt[i + 1]):
                        vertices.append((pts[j][0], pts[j][1]))
                    codes += [Path.MOVETO]
                    codes += [Path.LINETO] * (prt[i + 1] - prt[i] - 2)
                    codes += [Path.CLOSEPOLY]
                path = Path(vertices, codes)
        return path
    except Exception as err:
        print(err)
        return None


def shp2clip(originfig, ax, shpfile, region):
    sf = shapefile.Reader(shpfile)
    vertices = []  # 这块是已经修改的地方
    codes = []  # 这块是已经修改的地方
    region_lower = []
    if isinstance(region,list):
        for strs  in region:
            region_lower.append(strs.lower())
    else:
        region_lower.append(region.lower())
    for shape_rec in sf.shapeRecords():
        # if shape_rec.record[3] == region:  # 这里需要找到和region匹配的唯一标识符，record[]中必有一项是对应的。

        #if(len(shape_rec.record) ==1):continue
        if shape_rec.record[3].lower() in region_lower:  # 这块是已经修改的地方
            pts = shape_rec.shape.points
            prt = list(shape_rec.shape.parts) + [len(pts)]
            for i in range(len(prt) - 1):
                for j in range(prt[i], prt[i + 1]):
                    vertices.append((pts[j][0], pts[j][1]))
                codes += [Path.MOVETO]
                codes += [Path.LINETO] * (prt[i + 1] - prt[i] - 2)
                codes += [Path.CLOSEPOLY]
            path = Path(vertices, codes)
            # extents = path.get_extents()
            patch = PathPatch(path, transform=ax.transData, facecolor='none', edgecolor='black')
    for contour in originfig.collections:
        contour.set_clip_path(patch)
    return path, patch



def shp2clip_pro_id(originfig, ax, shpfile, num_list):
    sf = shapefile.Reader(shpfile)
    vertices = []  # 这块是已经修改的地方
    codes = []  # 这块是已经修改的地方

    shape_recs = sf.shapeRecords()

    for i in range(len(shape_recs)):
        # if shape_rec.record[3] == region:  # 这里需要找到和region匹配的唯一标识符，record[]中必有一项是对应的。

        #if(len(shape_rec.record) ==1):continue

        if i in num_list:  # 这块是已经修改的地方
            shape_rec = shape_recs[i]
            pts = shape_rec.shape.points
            prt = list(shape_rec.shape.parts) + [len(pts)]
            for i in range(len(prt) - 1):
                for j in range(prt[i], prt[i + 1]):
                    vertices.append((pts[j][0], pts[j][1]))
                codes += [Path.MOVETO]
                codes += [Path.LINETO] * (prt[i + 1] - prt[i] - 2)
                codes += [Path.CLOSEPOLY]
            path = Path(vertices, codes)
            # extents = path.get_extents()
            patch = PathPatch(path, transform=ax.transData, facecolor='none', edgecolor='black')
    for contour in originfig.collections:
        contour.set_clip_path(patch)
    return path, patch




def shp2clip_by_lines(originfig, ax, line_list):
    vertices = []  # 这块是已经修改的地方
    codes = []  # 这块是已经修改的地方
    for i in range(len(line_list)):
        # if shape_rec.record[3] == region:  # 这里需要找到和region匹配的唯一标识符，record[]中必有一项是对应的。

        #if(len(shape_rec.record) ==1):continue
        pts = line_list[i]
        for j in range(len(pts)):
            vertices.append((pts[j][0], pts[j][1]))
        vertices.append((pts[0][0], pts[0][1]))
        codes += [Path.MOVETO]
        codes += [Path.LINETO] * (len(pts)-1)
        codes += [Path.CLOSEPOLY]
    path = Path(vertices, codes)
        # extents = path.get_extents()
    patch = PathPatch(path, transform=ax.transData, facecolor='none', edgecolor='black')
    for contour in originfig.collections:
        contour.set_clip_path(patch)
    return path, patch


def shp2clip_by_region_name(originfig, ax, region_name_list):
    province = pkg_resources.resource_filename('meteva', "resources/maps/Province")

    province_ch_name = ["北京","天津","河北","山西","内蒙古",
                        "辽宁","吉林","黑龙江","上海","江苏",
                        "浙江","安徽","福建","江西", "山东",
                        "河南","湖北","湖南","广东","广西",
                        "海南","重庆","四川","贵州","云南",
                        "西藏","陕西","甘肃","青海","宁夏",
                        "新疆","台湾","香港","澳门"]

    province_name =["beijing","tianjin","hebei","shanxi","neimenggu",
                    "liaoning", "jilin", "heilongjiang", "shanghai", "jiangsu",
                    "zhejiang", "anhui", "fujian", "jiangxi", "shandong",
                    "henan", "hubei", "hunan", "guangdong", "guangxi",
                    "hainan", "chongqing", "sichuan", "guizhou", "yunnan",
                    "xizang", "shaanxi", "gansu", "qinghai", "ningxia",
                    "xinjiang", "taiwan", "xianggang", "aomen"]


    region_id_list = []
    #print(region_name_list)
    for region_name in region_name_list:
        if region_name.lower()=="china" or region_name=="中国":
            region_id_list.extend(np.arange(34).tolist())
        else:
            for i in range(len(province_name)):
                if region_name.find(province_ch_name[i])>=0:
                    region_id_list.append(i)
                if region_name.lower().find(province_name[i])>=0:
                    region_id_list.append(i)

    #print(region_id_list)
    region_id_list = list(set(region_id_list))
    shp2clip_pro_id(originfig, ax, province, region_id_list)


def shp2clip_by_shpfile(originfig, ax, shpfile):
    sf = shapefile.Reader(shpfile)
    vertices = []  # 这块是已经修改的地方
    codes = []  # 这块是已经修改的地方
    shape_recs = sf.shapeRecords()
    for i in range(len(shape_recs)):
        # if shape_rec.record[3] == region:  # 这里需要找到和region匹配的唯一标识符，record[]中必有一项是对应的。

        # if(len(shape_rec.record) ==1):continue
            shape_rec = shape_recs[i]
            pts = shape_rec.shape.points
            prt = list(shape_rec.shape.parts) + [len(pts)]
            for i in range(len(prt) - 1):
                for j in range(prt[i], prt[i + 1]):
                    vertices.append((pts[j][0], pts[j][1]))
                codes += [Path.MOVETO]
                codes += [Path.LINETO] * (prt[i + 1] - prt[i] - 2)
                codes += [Path.CLOSEPOLY]
            path = Path(vertices, codes)
            # extents = path.get_extents()
            patch = PathPatch(path, transform=ax.transData, facecolor='none', edgecolor='black')
    for contour in originfig.collections:
        contour.set_clip_path(patch)
    return path, patch





from . import score
from . import table
from . import plot
from .score import cr,crps,variance_mse,variance_divide_by_mse,std_rmse
from .table import *
from .plot import box_plot_ensemble,rank_histogram
